# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-29 18:14
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flickr', '0009_photoset_fetch_time'),
    ]

    operations = [
        migrations.AddField(
            model_name='photo',
            name='photosets',
            field=models.ManyToManyField(to='flickr.Photoset'),
        ),
    ]



from django.core.management.base import BaseCommand, CommandError


class DittoBaseCommand(BaseCommand):
    """
    Should probably be used as a parent by all other management commands,
    assuming they follow the usual pattern.

    Provides an output_results() method for our usual format of results.
    """

    # What we're fetching:
    singular_noun = 'Thing'
    plural_noun = 'Things'

    def add_arguments(self, parser):
        "We may add stuff for handling verbosity here."
        pass

    def output_results(self, results, verbosity=1):
        """results should be a list of dicts.

        Each dict is for one account.
        Each dict will look like either:

          { 'account': 'Phil Gyford',
            'success': True,
            'fetched': 200, # The number of things fetched
          }

        or:

          { 'account': 'Phil Gyford',
            'success': False,
            'messages': ['There was an error fetching data because blah'],
          }

        Or, it might not have an 'account'.
        """
        if verbosity == 0:
            return

        for result in results:
            prefix = '%s: ' % result['account'] if 'account' in result else ''
            if 'fetched' in result:
                noun = self.singular_noun if result['fetched'] == 1 else self.plural_noun
                self.stdout.write('%sFetched %s %s' % (
                                            prefix, result['fetched'], noun))

            if result['success'] == False:
                self.stderr.write('%sFailed to fetch %s: %s' % (
                                    prefix,
                                    self.plural_noun,
                                    self.format_messages(result['messages'])
                                ))

    def format_messages(self, messages):
        if len(messages) == 1:
            return messages[0]
        else:
            # On separate lines, and start a newline first.
            return "%s%s" % ("\n", "\n".join(messages))





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0022_auto_20151104_0840'),
    ]

    operations = [
        migrations.CreateModel(
            name='Media',
            fields=[
                ('id', models.AutoField(primary_key=True, auto_created=True, serialize=False, verbose_name='ID')),
                ('time_created', models.DateTimeField(help_text='The time this item was created in the database.', auto_now_add=True)),
                ('time_modified', models.DateTimeField(help_text='The time this item was last saved to the database.', auto_now=True)),
                ('media_type', models.CharField(choices=[('photo', 'Photo'), ('video', 'Video')], max_length=8)),
                ('twitter_id', models.BigIntegerField(unique=True)),
                ('image_url', models.URLField(help_text='URL of the image itself')),
                ('is_private', models.BooleanField(help_text='If true, this item will not be shown on public-facing pages.', default=False)),
                ('large_w', models.PositiveSmallIntegerField(null=True, verbose_name='Large width', blank=True)),
                ('large_h', models.PositiveSmallIntegerField(null=True, verbose_name='Large height', blank=True)),
                ('medium_w', models.PositiveSmallIntegerField(null=True, verbose_name='Medium width', blank=True)),
                ('medium_h', models.PositiveSmallIntegerField(null=True, verbose_name='Medium height', blank=True)),
                ('small_w', models.PositiveSmallIntegerField(null=True, verbose_name='Small width', blank=True)),
                ('small_h', models.PositiveSmallIntegerField(null=True, verbose_name='Small height', blank=True)),
                ('thumb_w', models.PositiveSmallIntegerField(null=True, verbose_name='Thumbnail width', blank=True)),
                ('thumb_h', models.PositiveSmallIntegerField(null=True, verbose_name='Thumbnail height', blank=True)),
                ('mp4_url_1', models.URLField(null=True, verbose_name='MP4 URL (1)', blank=True)),
                ('mp4_url_2', models.URLField(null=True, verbose_name='MP4 URL (2)', blank=True)),
                ('mp4_url_3', models.URLField(null=True, verbose_name='MP4 URL (3)', blank=True)),
                ('mp4_bitrate_1', models.PositiveIntegerField(null=True, verbose_name='MP4 Bitrate (1)', blank=True)),
                ('mp4_bitrate_2', models.PositiveIntegerField(null=True, verbose_name='MP4 Bitrate (2)', blank=True)),
                ('mp4_bitrate_3', models.PositiveIntegerField(null=True, verbose_name='MP4 Bitrate (3)', blank=True)),
                ('webm_url', models.URLField(null=True, verbose_name='WebM URL', blank=True)),
                ('aspect_ratio', models.CharField(help_text='eg, "4:3" or "16:9"', max_length=5)),
                ('duration', models.PositiveIntegerField(help_text='In milliseconds', null=True, blank=True)),
                ('tweet', models.ForeignKey(to='twitter.Tweet', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['time_created'],
            },
        ),
    ]



# coding: utf-8
import os

try:
    from django.urls import reverse
except ImportError:
    # For Django 1.8
    from django.urls import reverse
from django.db import models
from django.templatetags.static import static

from imagekit.cachefiles import ImageCacheFile

from . import app_settings
from . import imagegenerators
from . import managers
from .utils import htmlify_description, htmlify_tweet
from ..core.managers import PublicItemManager
from ..core.models import DiffModelMixin, DittoItemModel, TimeStampedModelMixin

import json


class Account(TimeStampedModelMixin, models.Model):
    """The Twitter User Accounts with which we fetch data from the API.
    Each one is connected to a User object, so we only need to store API
    details here.
    """

    user = models.ForeignKey('User', blank=True, null=True,
                                                    on_delete=models.SET_NULL)

    consumer_key = models.CharField(null=False, blank=True, max_length=255,
            help_text="(API Key) From https://apps.twitter.com")
    consumer_secret = models.CharField(null=False, blank=True, max_length=255,
            help_text="(API Secret)")
    access_token = models.CharField(null=False, blank=True, max_length=255)
    access_token_secret = models.CharField(null=False, blank=True,
                                                                max_length=255)
    last_recent_id = models.BigIntegerField(null=True, blank=True,
            help_text="The Twitter ID of the most recent Tweet fetched.")
    last_favorite_id = models.BigIntegerField(null=True, blank=True,
            help_text="The Twitter ID of the most recent favorited Tweet fetched.")

    is_active = models.BooleanField(default=True, null=False, blank=False,
                        help_text="If false, new Tweets won't be fetched.")

    def save(self, *args, **kwargs):
        if self.user is None:
            result = self.updateUserFromTwitter()
            # Quietly ignoring errors. Sorry.
        super().save(*args, **kwargs)

    def __str__(self):
        if self.user:
            return str(self.user)
        else:
            return '%d' % self.pk

    class Meta:
        ordering = ['user__screen_name']

    def get_absolute_url(self):
        if self.user:
            return reverse('twitter:user_detail',
                        kwargs={'screen_name': self.user.screen_name})
        else:
            return ''

    def updateUserFromTwitter(self):
        """Calls the Twitter API to fetch the user details for this account.
        If the user object doesn't exist yet, it is created.
        But its relationship with this Account isn't saved here, only assigned.

        Returns:
        a) A dict, or
        b) False if we don't have API credentials.

        If a dict, will be like one of:
        {'account': 'text string', 'success': False, 'message': 'Error msg'}
        or
        {'account': 'text string', 'success': True, 'user': <User obj>}
        """
        from .fetch.fetch import FetchVerify

        if self.has_credentials():
            results = FetchVerify(account=self).fetch()
            if 'user' in results and isinstance(results['user'], User):
                self.user = results['user']
            return results
        else:
            return False

    def has_credentials(self):
        "Does this at least have something in its API fields? True or False"
        if self.consumer_key and self.consumer_secret and self.access_token and self.access_token_secret:
            return True
        else:
            return False


class Media(TimeStampedModelMixin, models.Model):
    """A photo, video or animated GIF attached to a Tweet.

    They have a bunch of common fields, and then some extra for Videos.
    A Tweet could have zero, one or more Medias. Yes that's the plural shut up.
    """

    # Mapping our internal names for sizes to the imagekit generators:
    IMAGE_SIZES = {
        'medium':       { 'generator': imagegenerators.Medium, },
        'small':        { 'generator': imagegenerators.Small, },
        'thumb':        { 'generator': imagegenerators.Thumbnail, },
    }

    MEDIA_TYPES = (
        ('animated_gif', 'Animated GIF'),
        ('photo', 'Photo'),
        ('video', 'Video'),
    )

    media_type = models.CharField(null=False, blank=False, max_length=12,
                                                        choices=MEDIA_TYPES)

    # A media item can belong to more than one tweet. Why?
    # Because if tweet 1 retweets tweet 2, which has a photo, then we save
    # that photo as attached to both tweets 1 and 2.
    # Because that's how the API does it too.
    tweets = models.ManyToManyField('Tweet', related_name='media')

    twitter_id = models.BigIntegerField(null=False, blank=False, unique=True)
    image_url = models.URLField(null=False, blank=False,
                            help_text="URL of the image itself on Twitter.com")

    large_w = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Large width")
    large_h = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Large height")
    medium_w = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Medium width")
    medium_h = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Medium height")
    small_w = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Small width")
    small_h = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Small height")
    thumb_w = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Thumbnail width")
    thumb_h = models.PositiveSmallIntegerField(null=True, blank=True,
                                                verbose_name="Thumbnail height")

    # START VIDEO-ONLY PROPERTIES.

    # These will be in order from lowest bitrate to highest.
    mp4_url = models.URLField(null=False, blank=True,
                        verbose_name='MP4 URL', help_text="For Animated GIFs")
    dash_url = models.URLField(null=False, blank=True,
                                verbose_name='MPEG-DASH URL (.mpd, streaming)')
    xmpeg_url = models.URLField(null=False, blank=True,
                            verbose_name='X-MPEG URL (HLS, .m3u8, streaming')

    aspect_ratio = models.CharField(null=False, blank=True, max_length=21,
                                            help_text='eg, "4:3" or "16:9"')
    duration = models.PositiveIntegerField(null=True, blank=True,
                                                help_text="In milliseconds")
    # END VIDEO-ONLY PROPERTIES

    def upload_path(self, filename):
        "Make path under MEDIA_ROOT where original files will be saved."

        # eg get '12345678' from '12345678.jpg':
        name = os.path.splitext(filename)[0]

        # If filename is '12345678.jpg':
        # 'twitter/media/56/78/12345678.jpg'
        return '/'.join([
            app_settings.DITTO_TWITTER_DIR_BASE,
            'media',
            name[-4:-2],
            name[-2:],
            filename
        ])

    image_file = models.ImageField(upload_to=upload_path,
                                            null=False, blank=True, default='')
    mp4_file = models.FileField(upload_to=upload_path,
                                null=False, blank=True, default='',
                                help_text="Only used for Animated GIFs")

    # All Items (eg, used in Admin):
    objects = models.Manager()

    # There is no Public manager for media, as we don't have the concept of
    # private/public Media items.

    def __str__(self):
        return '%s %d' % (self.get_media_type_display(), self.id)

    class Meta:
        ordering = ['time_created']
        verbose_name = 'Media item'
        verbose_name_plural = 'Media items'

    @property
    def thumbnail_w(self):
        "Because we usually actually want 150, not whatever thumb_w is."
        return 150

    @property
    def thumbnail_h(self):
        "Because we usually actually want 150, not whatever thumb_h is."
        return 150

    @property
    def large_url(self):
        "URL to local or remote original-size image."
        return self._image_url('large')

    @property
    def medium_url(self):
        "URL to local or remote image, maximum 1200px high/wide."
        return self._image_url('medium')

    @property
    def small_url(self):
        "URL to local or remote image, maximum 680px high/wide."
        return self._image_url('small')

    @property
    def thumb_url(self):
        "URL to local or remote image, 150px square."
        return self._image_url('thumb')

    @property
    def thumbnail_url(self):
        "URL to local or remote image, 150px square. For consistency."
        return self.thumb_url

    @property
    def video_url(self):
        """
        The URL to our preferred video source.
        Depends on if we're hosting video locally or on twitter.com, what
        kind of thing this is (GIF or video), and what URLs we have.
        Returns a URL or an empty string.
        """
        url = ''
        video_type = self._video_type()

        if video_type is not None:
            if video_type[1] == 'local':
                if video_type[0] == 'mp4':
                    return self.mp4_file.url
            else:
                if video_type[0] == 'mp4':
                    url = self.mp4_url
                elif video_type[0] == 'xmpeg':
                    url = self.xmpeg_url
                elif video_type[0] == 'dash':
                    url = self.dash_url

        return url

    @property
    def video_mime_type(self):
        """
        The MIME type for our preferred video URL.
        Returns a mime type or an empty string.
        """
        mime_type = ''
        video_type = self._video_type()

        if video_type is not None:
            if video_type[0] == 'mp4':
                mime_type = 'video/mp4'
            elif video_type[0] == 'xmpeg':
                mime_type = 'application/x-mpegURL'
            elif video_type[0] == 'dash':
                mime_type = 'application/dash+xml'

        return mime_type

    def _image_url(self, size):
        """
        Helper for the self.*_url() property methods.
        size -- one of 'large', 'medium', 'small', or 'thumbnail'.
        """
        if app_settings.DITTO_TWITTER_USE_LOCAL_MEDIA:
            return self._local_image_url(size)
        else:
            return self._remote_image_url(size)

    def _local_image_url(self, size):
        """
        Generate the URL of an image of a particular size, hosted locally,
        based on the original file (which must already be downloaded).
        size -- one of 'large', 'medium', 'small', or 'thumbnail'.
        """
        if self.image_file:
            if size == 'large':
                # Essentially the original file.
                return self.image_file.url
            else:
                generator = self.IMAGE_SIZES[size]['generator']
                try:
                    image_generator = generator(source=self.image_file)
                    result = ImageCacheFile(image_generator)
                    return result.url
                except:
                    # We have an original file but something's wrong with it.
                    # Might be 0 bytes or something.
                    return static('img/original_error.jpg')
        else:
            # We haven't downloaded an original file for this.
            return static('img/original_missing.jpg')

    def _remote_image_url(self, size):
        """
        Generate the URL of an image of a particular size, at Twitter.
        size -- one of 'large', 'medium', 'small', or 'thumbnail'.
        """
        return  '%s:%s' % (self.image_url, size)

    def _video_type(self):
        """
        Which video type should we use for this item?
        Returns None or a tuple of type and 'remote'/'local',
        eg: ('xmpeg', 'remote') or ('mp4', 'local')
        """
        if app_settings.DITTO_TWITTER_USE_LOCAL_MEDIA:
            return self._local_video_type()
        else:
            return self._remote_video_type()

    def _local_video_type(self):
        """
        When hosting media locally, which video type should we use for this?
        Returns None or a tuple of type and 'remote'/'local',
        eg: ('xmpeg', 'remote') or ('mp4', 'local')
        """
        if self.media_type == 'animated_gif' and self.mp4_file:
            return ('mp4', 'local')
        else:
            # We can't currently save video files except MP4s for animated
            # gifs, so try to fall back to the remote URLs.
            return self._remote_video_type()

    def _remote_video_type(self):
        """
        When hosting media remotely, which video type should we use for this?
        Returns None or a tuple of type and 'remote'/'local',
        eg: ('xmpeg', 'remote') or ('mp4', 'local')
        """
        video_type = None
        if self.media_type == 'video':
            # Prefer the streaming URL over MP4:
            if self.xmpeg_url:
                video_type = ('xmpeg', 'remote')
            elif self.mp4_url:
                video_type = ('mp4', 'remote')
            elif self.dash_url:
                video_type = ('dash', 'remote')

        elif self.media_type == 'animated_gif' and self.mp4_url:
            video_type = ('mp4', 'remote')

        return video_type


class ExtraTweetManagers(models.Model):
    """Managers to use in the Tweet model, in addition to the defaults defined
    in DittoItemModel.
    These need to be here, rather than in the Tweet model, or they will
    override those in DittoItemModel.
    """
    # For Tweets favorited by any Account:
    favorite_objects = managers.FavoritesManager()
    public_favorite_objects = managers.PublicFavoritesManager()

    tweet_objects = managers.TweetsManager()
    public_tweet_objects = managers.PublicTweetsManager()

    class Meta:
        abstract = True


class Tweet(DittoItemModel, ExtraTweetManagers):
    """We don't replicate all of the possible Tweet attributes here, only
    enough to display the most useful things. Given we save the raw JSON
    about this tweet, we could add more attributes in future, even if original
    tweets are deleted on Twitter.

    Also, we don't include any 'perspectival' attributes - ones that will vary
    depending on which Account fetched this data. eg, `favorited` or
    `current_user_retweet`.
    """

    ditto_item_name = 'twitter_tweet'

    # Properties inherited from DittoItemModel:
    #
    # title         (CharField)
    # permalink     (URLField)
    # summary       (CharField)
    # is_private    (BooleanField)
    # fetch_time    (DateTimeField, UTC)
    # post_time     (DateTimeField, UTC)
    # latitude      (DecimalField)
    # longitude     (DecimalField)
    # raw           (TextField)

    user = models.ForeignKey('User', on_delete=models.CASCADE)

    text = models.TextField(null=False, blank=False)
    text_html = models.TextField(null=False, blank=True,
        help_text="An HTMLified version of the Tweet's text")
    twitter_id = models.BigIntegerField(null=False, blank=False, unique=True,
                                                                db_index=True)

    # Favorite and Retweet Count not present in tweets ingested from a
    # downloaded archive, hence null=True:
    favorite_count = models.PositiveIntegerField(null=True, blank=True,
        help_text="Approximately how many times this had been favorited when fetched")
    retweet_count = models.PositiveIntegerField(null=True, blank=True,
        help_text="Number of times this had been retweeted when fetched")

    in_reply_to_screen_name = models.CharField(null=False, blank=True,
        max_length=20,
        help_text="Screen name of the original Tweet's author, if this is a reply")
    in_reply_to_status_id = models.BigIntegerField(null=True, blank=True,
        help_text="The ID of the Tweet replied to, if any")
    in_reply_to_user_id = models.BigIntegerField(null=True, blank=True,
        help_text="ID of the original Tweet's author, if this is a reply")

    language = models.CharField(null=False, blank=False, default='und',
        max_length=20,
        help_text="A BCP 47 language identifier, or 'und' if it couldn't be detected")

    # Just enough of the place data to display something readable and useful:
    place_attribute_street_address = models.CharField(null=False, blank=True,
                                                                max_length=255)
    place_full_name = models.CharField(null=False, blank=True, max_length=255)
    place_country = models.CharField(null=False, blank=True, max_length=255)

    quoted_status_id = models.BigIntegerField(null=True, blank=True,
        help_text="The ID of the Tweet quoted, if any")

    retweeted_status_id = models.BigIntegerField(null=True, blank=True,
        help_text="The ID of the retweeted Tweet, if any")

    source = models.CharField(null=False, blank=True, max_length=255,
                                help_text="Utility used to post the Tweet")
    media_count = models.PositiveSmallIntegerField(null=False, blank=True,
        default=0, help_text="Number of Photos/Videos attached to this Tweet")

    def __str__(self):
        return self.title

    class Meta:
        ordering = ['-post_time']

    def save(self, *args, **kwargs):
        "Privacy depends on the user, so ensure it's set correctly"
        self.is_private = self.user.is_private
        result = self.make_text_html()
        super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse('twitter:tweet_detail', kwargs={
                                        'screen_name': self.user.screen_name,
                                        'twitter_id': self.twitter_id})

    def get_next_public_by_post_time(self):
        "Next Tweet by this User, if they're public."
        try:
            return Tweet.public_tweet_objects.filter(post_time__gte=self.post_time, user=self.user).exclude(pk=self.pk).order_by('post_time')[:1].get()
        except:
            pass

    def get_previous_public_by_post_time(self):
        "Previous Tweet by this User, if they're public."
        try:
            return Tweet.public_tweet_objects.filter(post_time__lte=self.post_time, user=self.user).exclude(pk=self.pk).order_by('-post_time')[:1].get()
        except:
            pass

    # Shortcuts:
    def get_next(self):
        return self.get_next_public_by_post_time()

    def get_previous(self):
        return self.get_previous_public_by_post_time()

    @property
    def is_reply(self):
        if self.in_reply_to_screen_name == '':
            return False
        else:
            return True

    @property
    def in_reply_to_url(self):
        "If it's a reply, the link to the tweet replied to."
        if self.is_reply:
            return 'https://twitter.com/%s/status/%s' % (
                    self.in_reply_to_screen_name, self.in_reply_to_status_id)
        else:
            return ''

    @property
    def place(self):
        return ', '.join(filter(None, (
                self.place_attribute_street_address,
                self.place_full_name,
                self.place_country,
            )))

    @property
    def account(self):
        "The Account whose tweet this is, if any. Otherwise, None."
        try:
            return self.user.account_set.all()[0]
        except IndexError:
            return None

    def make_text_html(self):
        """Uses the raw JSON for the tweet to set self.text_html to a nice
        HTML version of the tweet.
        """
        try:
            json_data = json.loads(self.raw)
        except ValueError as error:
            return False
        self.text_html = htmlify_tweet(json_data)
        return True

    def get_quoted_tweet(self):
        # There's one we saved earlier, so use that.
        if hasattr(self, '_quoted_tweet'):
            return self._quoted_tweet

        tweet = None
        if self.quoted_status_id:
            try:
                tweet = Tweet.public_objects.get(
                                            twitter_id=self.quoted_status_id)
            except Tweet.DoesNotExist:
                pass

        # Save for later:
        self._quoted_tweet = tweet
        return tweet

    def get_retweeted_tweet(self):
        # There's one we saved earlier, so use that.
        if hasattr(self, '_retweeted_tweet'):
            return self._retweeted_tweet

        tweet = None
        if self.retweeted_status_id:
            try:
                tweet = Tweet.public_objects.get(
                                        twitter_id=self.retweeted_status_id)
            except Tweet.DoesNotExist:
                pass

        # Save for later:
        self._retweeted_tweet = tweet
        return tweet

    def _summary_source(self):
        "Used to make the `summary` property."
        return self.title


class User(TimeStampedModelMixin, DiffModelMixin, models.Model):
    """A Twitter user.
    We don't replicate all of the possible User attributes here, only enough
    to display the most useful things.
    """

    twitter_id = models.BigIntegerField(null=False, blank=False, unique=True)
    screen_name = models.CharField(null=False, blank=False, max_length=20,
        help_text="Username, eg, 'samuelpepys'")
    name = models.CharField(null=False, blank=False, max_length=50,
        help_text="eg, 'Samuel Pepys'")
    url = models.URLField(null=False, blank=True, default='', max_length=255,
        help_text="A URL provided by the user as part of their profile")
    # Inverse of Twitter's 'protected', to be similar to DittoItemModel:
    is_private = models.BooleanField(null=False, default=False,
        help_text="True if this user is 'protected'")
    is_verified = models.BooleanField(null=False, default=False)

    # User data in ingested archives don't have this:
    created_at = models.DateTimeField(null=True, blank=True,
        help_text="UTC time when this account was created on Twitter")

    description = models.CharField(null=False, blank=True, max_length=255)
    description_html = models.TextField(null=False, blank=True,
                help_text="An HTMLified version of the User's description")
    location = models.CharField(null=False, blank=True, max_length=255)
    time_zone = models.CharField(null=False, blank=True, max_length=255)

    profile_image_url_https = models.URLField(null=False, blank=True,
                                                                max_length=255)

    # This is how it's spelt in the API:
    favourites_count = models.PositiveIntegerField(null=False, blank=False,
        default=0,
        help_text="The number of tweets this user has favorited in the account’s lifetime")
    followers_count = models.PositiveIntegerField(null=False, blank=False,
        default=0,
        help_text="The number of followers this account has")
    friends_count = models.PositiveIntegerField(null=False, blank=False,
        default=0,
        help_text="Tne number of users this account is following.")
    listed_count = models.PositiveIntegerField(null=False, blank=False,
        default=0,
        help_text="The number of public lists this user is a member of")
    statuses_count = models.PositiveIntegerField(null=False, blank=False,
        default=0,
        help_text="The number of tweets, including retweets, by this user")

    # As on DittoItemModel:
    fetch_time = models.DateTimeField(null=True, blank=True,
                            help_text="The time the data was last fetched.")
    raw = models.TextField(null=False, blank=True,
                                    help_text="eg, the raw JSON from the API.")

    def avatar_upload_path(self, filename):
        """
        Make path under MEDIA_ROOT where avatar file will be saved.

        eg, if twitter_id is 12345678:
            twitter/avatars/56/78/12345678/avatar_name.jpg
        """

        # We can't just join all these parts in one go, because if the ID
        # isn't long enough to have two numbered directories, (ie, it's only
        # 1 or 2 digits) then Django 1.8 creates a path with '//' rather than
        # just ignoring that directory.

        # So this is a bit laborious:

        # 'twitter/avatars':
        start = '/'.join([
            app_settings.DITTO_TWITTER_DIR_BASE,
            'avatars'
        ])
        # '/78/12345678/avatar_name.jpg':
        end = '/'.join([
            str(self.twitter_id)[-2:],
            str(self.twitter_id),
            str(filename)
        ])
        # The bit that will be empty for 1-2 digit IDs:
        # '56':
        middle = str(self.twitter_id)[-4:-2]

        if middle:
            return '/'.join([start, middle, end])
        else:
            return '/'.join([start, end])

    avatar = models.ImageField(upload_to=avatar_upload_path,
                                            null=False, blank=True, default='')

    favorites = models.ManyToManyField(Tweet, related_name="favoriting_users")

    objects = models.Manager()
    # All Users that have Accounts:
    objects_with_accounts = managers.WithAccountsManager()

    def __str__(self):
        return '@%s' % self.screen_name

    class Meta:
        ordering = ['screen_name']

    def save(self, *args, **kwargs):
        """If the user's privacy status has changed, we need to change the
        privacy of all their tweets
        And we also HTMLify their description.
        """
        if self.get_field_diff('is_private') is not None:
            Tweet.objects.filter(user=self).update(is_private=self.is_private)
        result = self.make_description_html()
        super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse('twitter:user_detail',
                    kwargs={'screen_name': self.screen_name})

    def make_description_html(self):
        """Uses the raw JSON for the user to set self.description_html to a nice
        HTML version of the description.
        """
        try:
            json_data = json.loads(self.raw)
        except ValueError as error:
            return False
        self.description_html = htmlify_description(json_data)
        return True

    @property
    def permalink(self):
        return 'https://twitter.com/%s' % self.screen_name

    @property
    def avatar_url(self):
        try:
            return self.avatar.url
        except ValueError:
            return static('img/default_avatar.png')

    @property
    def profile_image_url(self):
        return self.profile_image_url_https

    @property
    def favorites_count(self):
        return self.favourites_count




import calendar
import datetime
import os
import time

import flickrapi
from flickrapi.exceptions import FlickrError

from django.core.files import File

from . import FetchError
from .savers import UserSaver, PhotoSaver, PhotosetSaver
from ..models import Account, Photo, Photoset, User
from ...core.utils import datetime_now
from ...core.utils.downloader import DownloadException, filedownloader

# These classes call the Flickr API to fetch data about particular things,
# from the point of view of a single Account. eg, Photos, Users, Photosets.
#
# This data is then passed on to *Saver classes for creation/updating of
# objects.

# CLASSES HERE:
#
# Fetcher
#   UserIdFetcher
#   UserFetcher
#   PhotosFetcher
#       RecentPhotosFetcher
#   PhotosetsFetcher


class Fetcher(object):
    """Parent class for children that will call the Flickr API to fetch data.

    Depending on the child classes, it would be used something like:

        results = Fetcher(account=account_object).fetch()

    results is a dict that will have:
        'success': Boolean.
        'account': String. Indicating the Account (eg, its User's username).
        'fetched': Integer. If success, the number of things fetched, if any.
        'messages': List. If no success, the failure message(s).
    """

    # How many photos (or whatever) do we fetch per page of retults?
    items_per_page = 500

    def __init__(self, account):
        # Will be an Account object, passed into init()
        self.account = None

        # Will be the FlickrAPI object for calling the Flickr API.
        self.api = None

        # Will be incremented if we're fetching multiple pages.
        self.page_number = 1

        # If there are multiple pages to fetch, this should be set when we get
        # the first page.
        self.total_pages = 1

        # Will be the minimum date of upload/fave for Photos that we'll fetch.
        # By default, set it before Flickr so we get everything.
        self.min_date = datetime.datetime.strptime('2000-01-01', '%Y-%m-%d')

        # Will be the results fetched from the API via FlickrAPI.
        self.results = []

        # When fetching Photos or Users this will be the total number fetched.
        self.results_count = 0

        # What we'll return:
        self.return_value = {'fetched': 0}

        if isinstance(account, Account):
            if account.user:
                self.return_value['account'] = account.user.username
            elif account.pk:
                self.return_value['account'] = 'Account: %s' % str(account)
            else:
                self.return_value['account'] = 'Unsaved Account'
        else:
            raise ValueError("An Account object is required")

        if account.has_credentials():
            self.account = account
            self.api = flickrapi.FlickrAPI(self.account.api_key,
                                self.account.api_secret, format='parsed-json')
        else:
            self.return_value['success'] = False
            self.return_value['messages'] = ['Account has no API credentials']

    def fetch(self, **kwargs):
        self._fetch_pages(**kwargs)

        if self._not_failed():
            self.return_value['success'] = True
            self.return_value['fetched'] = self.results_count

        return self.return_value

    def _fetch_pages(self, **kwargs):
        """Fetch all of the pages.
        This relies on something else setting self.total_pages to the correct
        value. Probably in _call_api(), set it when fetching the first page,
        as the results will contain the total number of pages available.
        """
        while self.page_number <= self.total_pages and self._not_failed():
            self._fetch_page(**kwargs)
            self.page_number += 1
            time.sleep(0.5) # Being nice.

    def _fetch_page(self, **kwargs):
        try:
            self._call_api(**kwargs)
        except FetchError as e:
            self.return_value['success'] = False
            self.return_value['messages'] = [
                                    'Error when calling Flickr API: %s' % e]
            return

        try:
            self._fetch_extra()
        except FetchError as e:
            self.return_value['success'] = False
            self.return_value['messages'] = [
                                    'Error when fetching extra data: %s' % e]
            return

        try:
            self._save_results()
            # Clear for the next page:
            self.results = []
        except FetchError as e:
            self.return_value['success'] = False
            self.return_value['messages'] = ['Error when saving data: %s' % e]
            return

        return

    def _not_failed(self):
        """Has everything gone smoothly so far? ie, no failure registered?"""
        if 'success' not in self.return_value or self.return_value['success'] == True:
            return True
        else:
            return False

    def _call_api(self, **kwargs):
        """
        Should call self.api.a_function() and set self.results with the results.
        """
        raise FetchError(
            "Subclasess of Fetcher should define their own _call_api().")

    def _fetch_extra(self):
        """Can be defined in subclasses to fetch extra data to add to
        self.results before we save the data in the DB."""
        pass

    def _save_results(self, **kwargs):
        """
        Should go through self.results and create/update things in the DB based
        on its contents.
        """
        raise FetchError(
            "Subclasses of Fetcher should define their own _save_results().")


class UserIdFetcher(Fetcher):
    """Returns the Flickr ID for an Account's user.
    Doesn't save anything.

    Use like:
        results = UserIdFetcher(account=account_obj).fetch()

    Results will look something like:
        {
            'id': '35034346050@N01',
            'fetched': 1,
            'account': 'Phil Gyford',
            'success': True,
        }
    """

    def _call_api(self):
        """Calls test.login() to get the very basic user info for the
        authenticating user.
        Docs: https://www.flickr.com/services/api/flickr.test.login.htm
        """
        try:
            info = self.api.test.login()
        except FlickrError as e:
            raise FetchError("Error when calling test.login(): %s'" % e)

        self.results = [ {'id': info['user']['id']} ]

    def _save_results(self):
        "Doesn't save any results, just prepares to return the Flickr ID."
        self.return_value['id'] = self.results[0]['id']
        self.results_count = 1


class UserFetcher(Fetcher):
    """Fetches and saves data about a single User by Flickr ID."""

    def fetch(self, nsid=None):
        """Fetches and saves data about a single User by Flickr ID.
        nsid -- A Flickr ID for a user.
        """
        if nsid is None:
            raise FetchError(
                        "UserFetcher().fetch() requires a Flickr id (NSID)")

        return super().fetch(nsid=nsid)

    def _call_api(self, nsid):
        "nsid -- A Flickr ID for a user."
        try:
            info = self.api.people.getInfo(user_id=nsid)
        except FlickrError as e:
            raise FetchError(
                "Error when getting info about User with Flickr ID '%s': %s" %\
                                                                    (nsid, e))

        # info has 'person' and 'stat' elements.
        self.results = [ info['person'] ]

    def _save_results(self):
        user_obj = UserSaver().save_user(self.results[0], datetime_now())
        self._fetch_and_save_avatar(user_obj)
        self.return_value['user'] = {'name': user_obj.name}
        self.results_count = 1

    def _fetch_and_save_avatar(self, user):
        """
        Download and save the Avatar/profile pic for this user.
        user -- User object.
        """
        try:
            avatar_filepath = filedownloader.download(user.original_icon_url,
                        ['image/jpeg', 'image/jpg', 'image/png', 'image/gif'])

            user.avatar.save(os.path.basename(avatar_filepath),
                                File(open(avatar_filepath, 'rb')) )
        except DownloadException as e:
            pass


class PhotosFetcher(Fetcher):
    """Parent class for fetching and saving data about Photos for an Account.
    """

    def __init__(self, *args, **kwargs):
        # Will match Flickr IDs with their User object.
        # eg '35034346050@N01' => User
        # When we fetch a user's data, because we need it to save a photo/tag,
        # we add their object to this, so we don't fetch again this time.
        self.fetched_users = {}

        super().__init__(*args, **kwargs)

    def _call_api(self):
        """
        Should call self.api.a_function() and set self.results with the results.
        """
        raise FetchError(
            "Subclasess of PhotosFetcher should define their own _call_api().")

    def _fetch_extra(self):
        """Before saving we need to go through the big list of photos we've
        fetched, and fetch more detailed info to add to each photo's data.
        """
        extra_results = []

        for i, photo in enumerate(self.results):

            self._fetch_user_if_missing(photo['owner'])

            extra_results.append({
                'fetch_time': datetime_now(),
                # Add the data for the photo's owner:
                'user_obj': self.fetched_users[ photo['owner'] ],
                # Get all the info about this photo:
                'info': self._fetch_photo_info(photo['id']),
                'sizes': self._fetch_photo_sizes(photo['id']),
                'exif': self._fetch_photo_exif(photo['id']),
            })

        # Replace self.results with our new array that contains more info.
        self.results = extra_results

    def _fetch_user_if_missing(self, flickr_user_id):
        """
        If we don't have flickr_user_id in self.fetched_users, then fetch, and
        save, that user from the API. Then add their User to self.fetched_users.
        flickr_user_id -- The user's ID on Flickr, eg '35034346050@N01'.
        """
        if self.fetched_users.get(flickr_user_id, None) is None:
            results = UserFetcher(account=self.account).fetch(
                                                        nsid=flickr_user_id)
            if results['success'] == False:
                raise FetchError(results['messages'][0])

            # Get the user we just saved. A bit clunky!
            self.fetched_users[flickr_user_id] = User.objects.get(
                                                           nsid=flickr_user_id)

    def _fetch_photo_info(self, photo_id):
        """Calls the photos.getInfo() method of the Flickr API and returns the
        info about the photo.
        https://www.flickr.com/services/api/explore/flickr.photos.getInfo
        photo_id -- The Flickr photo ID.
        """
        try:
            results = self.api.photos.getInfo(photo_id = photo_id)
        except FlickrError as e:
            raise FetchError("Error when fetching photo info (photo %s): %s" % \
                                                                (photo_id, e))

        # Each tag on the photo is added by a specific Flickr user.
        # (Usually, but not always, the photo owner.)
        # Check that we've got info about that user in our DB.
        for tag in results['photo']['tags']['tag']:
            self._fetch_user_if_missing(tag['author'])

        return results['photo']

    def _fetch_photo_sizes(self, photo_id):
        """Calls the photos.getSizes() method of the Flickr API and returns the
        photo's sizes.
        https://www.flickr.com/services/api/explore/flickr.photos.getSizes
        photo_id -- The Flickr photo ID.
        """
        try:
            results = self.api.photos.getSizes(photo_id = photo_id)
        except FlickrError as e:
            raise FetchError(
                "Error when fetching photo sizes (photo %s): %s" % \
                                                                (photo_id, e))
        return results['sizes']

    def _fetch_photo_exif(self, photo_id):
        """Calls the photos.getExif() method of the Flickr API and returns the
        photo's EXIF data.
        https://www.flickr.com/services/api/explore/flickr.photos.getExif
        photo_id -- The Flickr photo ID.
        """
        try:
            results = self.api.photos.getExif(photo_id = photo_id)
        except FlickrError as e:
            raise FetchError(
                "Error when fetching photo EXIF data (photo %s): %s" % \
                                                                (photo_id, e))
        return results['photo']

    def _save_results(self):
        """Save all the data we've fetched about photos to the DB."""
        saver = PhotoSaver()
        for photo in self.results:
            p = saver.save_photo(photo)
        self.results_count += len(self.results)


class RecentPhotosFetcher(PhotosFetcher):
    """Fetches and saves data about an Account's recent Photos.

    By default ALL photos will be fetched.

    Supply a number of days to fetch() to restrict to the most recent days.
    """

    def fetch(self, days=None):
        """Fetch all of the Account's user's photos, by default.
        days -- Required. The number of days back to look, by upload date, or
                'all' to fetch all photos.
        """
        try:
            self.min_date = datetime_now() - datetime.timedelta(days=days)
        except TypeError:
            if days != 'all':
                raise FetchError("days should be a number or 'all'.")

        return super().fetch()

    def _call_api(self):
        """Fetch one page of results, containing very basic info about the
        Photos."""

        # Turn our datetime object into a unix timestamp:
        min_unixtime = calendar.timegm(self.min_date.timetuple())

        try:
            results = self.api.people.getPhotos(
                                            user_id=self.account.user.nsid,
                                            min_upload_date=min_unixtime,
                                            per_page=self.items_per_page,
                                            page=self.page_number
                                        )
        except FlickrError as e:
            raise FetchError(
                "Error when fetching recent photos (page %s): %s" % \
                                                        (self.page_number, e))

        if self.page_number == 1 and 'photos' in results and 'pages' in results['photos']:
            # First time, set the total_pages there are to fetch.
            self.total_pages = int(results['photos']['pages'])

        # Add the list of photos' data from this page on to our total list:
        self.results += results['photos']['photo']


class PhotosetsFetcher(Fetcher):

    def _call_api(self):
        """Fetch one page of results.

        https://www.flickr.com/services/api/flickr.photosets.getList.htm
        """

        try:
            results = self.api.photosets.getList(
                                            user_id=self.account.user.nsid,
                                            per_page=self.items_per_page,
                                            page=self.page_number
                                        )
        except FlickrError as e:
            raise FetchError(
                "Error when fetching photosets (page %s): %s" % \
                                                        (self.page_number, e))

        if self.page_number == 1 and 'photosets' in results and 'pages' in results['photosets']:
            # First time, set the total_pages there are to fetch.
            self.total_pages = int(results['photosets']['pages'])

        # Add the list of photosets' data from this page on to our total list:
        self.results += results['photosets']['photoset']

    def _fetch_extra(self):
        """Before saving we need to get the list of photos in each photoset."""

        extra_results = []

        for i, photoset in enumerate(self.results):

            photos = self._fetch_photos_in_photoset(photoset['id'])

            extra_results.append({
                'fetch_time': datetime_now(),
                'photoset': photoset,
                'photos': photos,
                'user_obj': self.account.user,
            })

        # Replace self.results with our new array that contains more info.
        self.results = extra_results

    def _fetch_photos_in_photoset(self, photoset_id):
        """Gets the info about all the photos in the photoset.
        Might be in multiple pages.
        The data is quite minimal, but that's all we need.

        https://www.flickr.com/services/api/flickr.photosets.getPhotos.htm

        Expects:
            photoset_id -- eg 72157665648859705
        Returns:
            list of dicts, each dict containing photo data
        """

        photos = []
        page_number = 1
        total_pages = 1 # Will get set to its proper value below.

        while page_number <= total_pages:
            try:
                results = self.api.photosets.getPhotos(
                                                photoset_id=photoset_id,
                                                user_id=self.account.user.nsid,
                                                per_page=self.items_per_page,
                                                page=page_number
                                            )
            except FlickrError as e:
                raise FetchError(
                    "Error when fetching photos in photoset %s (page %s): %s"%\
                                                (photoset_id, page_number, e))

            if 'photoset' in results and 'photo' in results['photoset']:
                total_pages = results['photoset']['pages']
                photos += results['photoset']['photo']

            page_number += 1
            time.sleep(0.5) # Being nice.

        return photos

    def _save_results(self):
        """Save all the data we've fetched about photosets to the DB."""
        saver = PhotosetSaver()
        for photoset in self.results:
            p = saver.save_photoset(photoset)
        self.results_count += len(self.results)




# coding: utf-8
import argparse

from django.core.management.base import BaseCommand, CommandError

from ...fetch.fetchers import VerifyFetcher


class Command(BaseCommand):
    """Updates the stored data about the Twitter user for one or all Accounts.

    For one account:
    ./manage.py fetch_accounts --account=philgyford

    For all accounts:
    ./manage.py fetch_accounts
    """

    help = "Fetches and updates data about Accounts' Twitter Users"

    def add_arguments(self, parser):
        parser.add_argument(
            '--account',
            action='store',
            default=False,
            help='Only fetch for one Twitter account.',
        )

    def handle(self, *args, **options):
        # We might be fetching for a specific account or all (None).
        account = options['account'] if options['account'] else None;

        results = VerifyFetcher(screen_name=account).fetch()

        # results should be a list of dicts, either:
        # { 'account': 'thescreenname',
        #   'success': True
        # }
        # or:
        # { 'account': 'thescreenname',
        #   'success': False,
        #   'messages': ["This screen_name doesn't exist"]
        # }
        if options.get('verbosity', 1) > 0:
            for result in results:
                if result['success']:
                    self.stdout.write('Fetched @%s' % result['account'])
                else:
                    self.stderr.write('Could not fetch @%s: %s' % (
                                    result['account'], result['messages'][0]
                                ))




# coding: utf-8
from django.db import models
try:
    from django.urls import reverse
except ImportError:
    # For Django 1.8
    from django.urls import reverse

from ..core.models import DiffModelMixin, DittoItemModel, TimeStampedModelMixin
from ..core.utils import truncate_string
from . import managers


# For generating permalinks.
LASTFM_URL_ROOT = 'http://www.last.fm'

# For generating links to MB.
MUSICBRAINZ_URL_ROOT = 'https://musicbrainz.org'


class Account(TimeStampedModelMixin, models.Model):
    "The account of a user, with API key, on Last.fm"

    username = models.CharField(null=False, blank=False, max_length=30,
                unique=True,
                help_text="eg, 'rj'")

    realname = models.CharField(null=False, blank=False, max_length=30,
                unique=False,
                help_text="eg, 'Richard Jones'")

    api_key = models.CharField(blank=True, max_length=255,
                                                        verbose_name='API Key')
    is_active = models.BooleanField(default=True,
                        help_text="If false, new scrobbles won't be fetched.")

    def __str__(self):
        return self.realname

    class Meta:
        ordering = ['username']

    def get_absolute_url(self):
        return reverse('lastfm:user_detail',
                                        kwargs={'username': self.username})

    @property
    def permalink(self):
        return '%s/user/%s' % (LASTFM_URL_ROOT, self.username)

    def has_credentials(self):
        "Does this at least have something in its API field? True or False"
        if self.api_key:
            return True
        else:
            return False

    def get_recent_scrobbles(self, limit=10):
        return self.scrobbles.prefetch_related('artist', 'track')\
                                .order_by('-post_time')[:limit]


class Album(TimeStampedModelMixin, models.Model):
    """
    Minimal model of a music album.

    Last.fm's data isn't great - the album might be by various artists but,
    if the album has no MBID in the scrobble, we'll end up creating that album
    for each artist on the album.

    eg, with artist/track/album scrobbles like:
        Kelis / Milkshake / The Trevor Nelson Collection
        Luther Vandross / Never Too Much / The Trevor Nelson Collection

    we'd end up with two albums:
        Kelis / The Trevor Nelson Collection
        Luther Vandross / The Trevor Nelson Collection
    """

    name = models.TextField(null=False, blank=False)
    # We're not using SlugField because a Track slug can be longer than 255
    # characters and contain characters not allowed by SlugField().
    slug = models.TextField(null=False, blank=False, db_index=True,
            help_text="Lowercase")
    original_slug = models.TextField(null=False, blank=False,
            help_text="As used on Last.fm. Mixed case.")
    artist = models.ForeignKey('Artist', on_delete=models.CASCADE,
                                                        related_name='albums')
    mbid = models.CharField(null=False, blank=True, max_length=36,
            verbose_name="MBID",
            help_text="MusicBrainz Identifier")

    objects = managers.AlbumsManager()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def get_absolute_url(self):
        "The Album's URL locally."
        return reverse('lastfm:album_detail', kwargs={
            'artist_slug': self.artist.slug,
            'album_slug': self.slug,
        })

    @property
    def permalink(self):
        "The Album's URL at Last.fm."
        return '%s/music/%s/%s' % (LASTFM_URL_ROOT,
                                   self.artist.original_slug,
                                   self.original_slug)

    @property
    def musicbrainz_url(self):
        if self.mbid:
            return '%s/release/%s' % (MUSICBRAINZ_URL_ROOT, self.mbid)
        else:
            return None

    @property
    def tracks(self):
        """
        A QuerySet of all the Tracks on this Album, ordered by how many times
        they were scrobbled, most-scrobbled first.
        """
        qs = Track.objects.with_scrobble_counts(album=self)
        return qs

    def get_scrobble_count(self):
        """
        If we just have a `scrobble_count` property it clashes when we use
        the Album.objects.with_scrobble_count() query.
        """
        return self.scrobbles.count()

    def get_most_recent_scrobble(self):
        """
        Returns the most recent Scrobble object for this Album.
        """
        return self.scrobbles.order_by('-post_time').first()


class Artist(TimeStampedModelMixin, models.Model):
    "Minimal model of a music artist."

    name = models.CharField(null=False, blank=False, max_length=255)
    # We're not using SlugField because an Artist slug can be longer than 255
    # characters and contain characters not allowed by SlugField().
    slug = models.TextField(null=False, blank=False, unique=True,
            help_text="Lowercase")
    original_slug = models.TextField(null=False, blank=False, unique=True,
            help_text="As used on Last.fm. Mixed case.")
    mbid = models.CharField(null=False, blank=True, max_length=36,
            verbose_name="MBID",
            help_text="MusicBrainz Identifier")

    objects = managers.ArtistsManager()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def get_absolute_url(self):
        "The Artist's URL locally."
        return reverse('lastfm:artist_detail', kwargs={
            'artist_slug': self.slug,
        })

    @property
    def permalink(self):
        "The Artist's URL at Last.fm."
        return '%s/music/%s' % (LASTFM_URL_ROOT, self.original_slug)

    @property
    def musicbrainz_url(self):
        if self.mbid:
            return '%s/artist/%s' % (MUSICBRAINZ_URL_ROOT, self.mbid)
        else:
            return None

    def get_scrobble_count(self):
        """
        If we just have a `scrobble_count` property it clashes when we use
        the Artist.objects.with_scrobble_count() query.
        """
        return self.scrobbles.count()

    def get_top_albums(self, limit='all'):
        """
        Returns a QuerySet of Albums by this Artist, ordered by most-scrobbled.
        By default returns all of them.
        `limit` is 'all' for all albums, or an integer to return that number.
        """
        qs = self.albums.with_scrobble_counts()

        if limit == 'all':
            return qs
        else:
            return qs[:limit]

    def get_top_tracks(self, limit='all'):
        """
        Returns a QuerySet of Tracks by this Artist, ordered by most-scrobbled.
        By default returns all of them.
        `limit` is 'all' for all tracks, or an integer to return that number.
        """
        qs = self.tracks.with_scrobble_counts()

        if limit == 'all':
            return qs
        else:
            return qs[:limit]

    def get_most_recent_scrobble(self):
        """
        Returns the most recent Scrobble object for this Artist.
        """
        return self.scrobbles.order_by('-post_time').first()


class Scrobble(DittoItemModel, models.Model):
    """
    A single play of a single music track.

    A lot of scrobbles have an album title, but the album has no MBID.
    In that case we record the title, but don't link this Scrobble to an
    Album model.
    """

    ditto_item_name = 'lastfm_scrobble'

    # Properties inherited from DittoItemModel:
    #
    # title         (CharField)
    # permalink     (URLField)
    # summary       (CharField)
    # is_private    (BooleanField)
    # fetch_time    (DateTimeField, UTC)
    # post_time     (DateTimeField, UTC) (ie, time of scrobble)
    # latitude      (DecimalField)
    # longitude     (DecimalField)
    # raw           (TextField)

    account = models.ForeignKey('Account', on_delete=models.CASCADE,
                                                    related_name='scrobbles')

    artist = models.ForeignKey('Artist', on_delete=models.CASCADE,
                                                    related_name='scrobbles')
    track = models.ForeignKey('Track', on_delete=models.CASCADE,
                                                    related_name='scrobbles')
    album = models.ForeignKey('Album', on_delete=models.SET_NULL,
                            related_name='scrobbles', blank=True, null=True)

    def __str__(self):
        return '%s (%s)' % (self.title, self.post_time)

    class Meta:
        ordering = ['-post_time']

    def save(self, *args, **kwargs):
        self.title = truncate_string(
            '{} – {}'.format(self.track.artist.name, self.track.name),
            chars=255,
            truncate='…',
            at_word_boundary=True
        )
        super().save(*args, **kwargs)

    def _summary_source(self):
        "Used to make the `summary` property."
        return self.post_time.strftime('%Y-%m-%d %H:%M')


class Track(TimeStampedModelMixin, models.Model):
    """
    Minimal model of a music track.

    We don't link Tracks to Albums directly. We only record the album info
    in the Scrobble, as it all seems a bit dodgy in Last.fm's modelling.
    """

    name = models.TextField(null=False, blank=False)
    # We're not using SlugField because a Track slug can be longer than 255
    # characters and contain characters not allowed by SlugField().
    slug = models.TextField(null=False, blank=False, db_index=True,
            help_text="Lowercase")
    original_slug = models.TextField(null=False, blank=False,
            help_text="As used on Last.fm. Mixed case.")
    artist = models.ForeignKey('Artist', on_delete=models.CASCADE,
                                                        related_name='tracks')
    mbid = models.CharField(null=False, blank=True, max_length=36,
            verbose_name="MBID",
            help_text="MusicBrainz Identifier")

    objects = managers.TracksManager()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def get_absolute_url(self):
        "The track's URL locally."
        return reverse('lastfm:track_detail', kwargs={
            'artist_slug': self.artist.slug,
            'track_slug': self.slug,
        })

    @property
    def permalink(self):
        "The Track's URL at Last.fm."
        return '%s/music/%s/_/%s' % (LASTFM_URL_ROOT,
                                     self.artist.original_slug,
                                    self.original_slug)

    @property
    def musicbrainz_url(self):
        if self.mbid:
            return '%s/recording/%s' % (MUSICBRAINZ_URL_ROOT, self.mbid)
        else:
            return None

    @property
    def albums(self):
        """
        A QuerySet of all the Albums on which this Track has appeared, ordered
        by how many times the Album was scrobbled, most-scrobbled first.
        """
        return Album.objects.with_scrobble_counts(track=self)

    def get_scrobble_count(self):
        """
        If we just have a `scrobble_count` property it clashes when we use
        the Track.objects.with_scrobble_count() query.
        """
        return self.scrobbles.count()

    def get_most_recent_scrobble(self):
        """
        Returns the most recent Scrobble object for this Track.
        """
        return self.scrobbles.order_by('-post_time').first()




import argparse

from django.core.management.base import BaseCommand, CommandError

from ...fetch.fetchers import UserIdFetcher, UserFetcher
from ...models import Account, User


class Command(BaseCommand):
    """For fetching data about an Account's Flickr user.

    Should create/update them in our DB, and associate the User with the Account.

        ./manage.py fetch_flickr_account_user --id=1
    """

    help = "Fetches data for an Account's Flickr User"

    def add_arguments(self, parser):
        parser.add_argument(
            '--id',
            action='store',
            default=False,
            help='The ID of the Account in the Django database.',
            type=int
        )

    def handle(self, *args, **options):
        if options['id'] is False:
            raise CommandError("Specify an Account ID like --id=1")

        # First we need the Account object we're fetching for.
        account = False
        try:
            account = Account.objects.get(id=options['id'])
        except Account.DoesNotExist:
            self.stderr.write("No Account found with an id of '%s'" % options['id'])

        if account:
            # Then get the ID of the Flicker user for this Account's API creds.
            id_result = UserIdFetcher(account=account).fetch()

            if 'success' in id_result and id_result['success'] == True and 'id' in id_result:
                # We've got a Flickr ID, so we can get and save the user data.
                result = UserFetcher(account=account).fetch(
                                                        nsid=id_result['id'])
                if 'success' in result and result['success'] == True:
                    # Now we'll associate this User with the Account:
                    user = User.objects.get(nsid='35034346050@N01')
                    account.user = user
                    account.save()
                    if options.get('verbosity', 1) > 0:
                        self.stdout.write("Fetched and saved user '%s'" %\
                                                        result['user']['name'])
                else:
                    if options.get('verbosity', 1) > 0:
                        self.stderr.write(
                        "Failed to fetch a user using Flickr ID '%s': %s" %\
                                    (id_result['id'], result['messages'][0]))
            else:
                if options.get('verbosity', 1) > 0:
                    self.stderr.write(
                    "Failed to fetch a Flickr ID for this Account: %s" %\
                                                    id_result['messages'][0])





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0019_auto_20151028_1556'),
    ]

    operations = [
        migrations.RenameField(
            model_name='photo',
            old_name='media_url',
            new_name='url',
        ),
        migrations.RemoveField(
            model_name='photo',
            name='page_url',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0023_media'),
    ]

    operations = [
        migrations.AddField(
            model_name='media',
            name='dash_url',
            field=models.URLField(null=True, verbose_name='MPEG-DASH URL', blank=True),
        ),
        migrations.AddField(
            model_name='media',
            name='webm_bitrate',
            field=models.PositiveIntegerField(null=True, verbose_name='WebM Bitrate', blank=True),
        ),
        migrations.AddField(
            model_name='media',
            name='xmpeg_url',
            field=models.URLField(null=True, verbose_name='X-MPEG URL', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-28 15:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinboard', '0020_auto_20160428_1526'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bookmark',
            name='url_hash',
            field=models.CharField(blank=True, db_index=True, help_text="Slug in the Bookmark's local URL.", max_length=12, unique=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
import ditto.core.models


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0002_auto_20150729_1704'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tweet',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('time_created', models.DateTimeField(help_text=b'The time this item was created in the database.', auto_now_add=True)),
                ('time_modified', models.DateTimeField(help_text=b'The time this item was last saved to the database.', auto_now=True)),
                ('title', models.CharField(max_length=255, blank=True)),
                ('permalink', models.URLField(help_text=b"URL of the item on the service's website.", blank=True)),
                ('summary', models.CharField(help_text=b"eg, Initial text of a blog post, start of the description of a photo, all of a Tweet's text, etc. No HTML.", max_length=255, blank=True)),
                ('is_private', models.BooleanField(default=False, help_text=b'If set, this item will not be shown on public-facing pages.')),
                ('fetch_time', models.DateTimeField(help_text=b"The time the item's data was last fetched, and was new or changed.", null=True, blank=True)),
                ('latitude', models.DecimalField(null=True, max_digits=12, decimal_places=9, blank=True)),
                ('longitude', models.DecimalField(null=True, max_digits=12, decimal_places=9, blank=True)),
                ('raw', models.TextField(help_text=b'eg, the raw JSON from the API.', blank=True)),
                ('text', models.CharField(max_length=255)),
                ('twitter_id', models.BigIntegerField(unique=True)),
                ('twitter_id_str', models.CharField(unique=True, max_length=20)),
                ('created_at', models.DateTimeField(help_text=b'UTC time when this Tweet was created on Twitter')),
                ('favorite_count', models.PositiveIntegerField(help_text=b'Approximately how many times this has been favorited')),
                ('retweet_count', models.PositiveIntegerField(help_text=b'Number of times this has been retweeted')),
                ('in_reply_to_screen_name', models.CharField(help_text=b"Screen name of the original Tweet's author, if this is a reply", max_length=20, blank=True)),
                ('in_reply_to_status_id', models.BigIntegerField(help_text=b'The ID of the Tweet replied to, if any', null=True, blank=True)),
                ('in_reply_to_status_id_str', models.CharField(max_length=20, blank=True)),
                ('in_reply_to_user_id', models.BigIntegerField(help_text=b"ID of the original Tweet's author, if this is a reply", null=True, blank=True)),
                ('in_reply_to_user_id_str', models.CharField(max_length=20, blank=True)),
                ('language', models.CharField(default=b'und', help_text=b"A BCP 47 language identifier, or 'und' if it couldn't be detected", max_length=20)),
                ('place_attribute_street_address', models.CharField(max_length=255, blank=True)),
                ('place_full_name', models.CharField(max_length=255, blank=True)),
                ('place_country', models.CharField(max_length=255, blank=True)),
                ('quoted_status_id', models.BigIntegerField(help_text=b'The ID of the Tweet quoted, if any', null=True, blank=True)),
                ('quoted_status_id_str', models.CharField(max_length=20, blank=True)),
                ('source', models.CharField(help_text=b'Utility used to post the Tweet', max_length=255, blank=True)),
            ],
            options={
                'ordering': ['-created_at'],
            },
            bases=(ditto.core.models.DiffModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('time_created', models.DateTimeField(help_text=b'The time this item was created in the database.', auto_now_add=True)),
                ('time_modified', models.DateTimeField(help_text=b'The time this item was last saved to the database.', auto_now=True)),
                ('twitter_id', models.BigIntegerField(unique=True)),
                ('twitter_id_str', models.CharField(unique=True, max_length=20)),
                ('screen_name', models.CharField(help_text=b"Username, eg, 'samuelpepys'", max_length=20)),
                ('name', models.CharField(help_text=b"eg, 'Samuel Pepys'", max_length=30)),
                ('url', models.URLField(help_text=b'A URL provided by the user as part of their profile', blank=True)),
                ('is_private', models.BooleanField(default=False, help_text=b"True if this user is 'protected' or private")),
                ('is_verified', models.BooleanField(default=False)),
                ('created_at', models.DateTimeField(help_text=b'UTC time when this account was created on Twitter')),
                ('description', models.CharField(max_length=255, blank=True)),
                ('location', models.CharField(max_length=255, blank=True)),
                ('time_zone', models.CharField(max_length=255, blank=True)),
                ('profile_image_url', models.URLField(max_length=255, blank=True)),
                ('profile_image_url_https', models.URLField(max_length=255, blank=True)),
                ('favorites_count', models.PositiveIntegerField(help_text=b'The number of tweets this user has favorited in the account\xe2\x80\x99s lifetime')),
                ('followers_count', models.PositiveIntegerField(help_text=b'The number of followers this account has')),
                ('friends_count', models.PositiveIntegerField(help_text=b'Tne number of users this account is following.')),
                ('listed_count', models.PositiveIntegerField(help_text=b'The number of public lists this user is a member of')),
                ('statuses_count', models.PositiveIntegerField(help_text=b'The number of tweets, including retweets, by this user')),
                ('fetch_time', models.DateTimeField(help_text=b'The time the data was last fetched, and was new or changed.', null=True, blank=True)),
                ('raw', models.TextField(help_text=b'eg, the raw JSON from the API.', blank=True)),
            ],
            options={
                'ordering': ['screen_name'],
            },
            bases=(ditto.core.models.DiffModelMixin, models.Model),
        ),
        migrations.AlterField(
            model_name='account',
            name='screen_name',
            field=models.CharField(help_text=b"Username, eg, 'samuelpepys'", unique=True, max_length=20, blank=True),
        ),
        migrations.AddField(
            model_name='tweet',
            name='user',
            field=models.ForeignKey(to='twitter.User', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='account',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, blank=True, to='twitter.User', null=True),
        ),
    ]



from django.apps import AppConfig


class DittoFlickrConfig(AppConfig):
    name = 'ditto.flickr'
    verbose_name = "Ditto Flickr"

    def ready(self):
        # import ditto.flickr.signals
        import ditto.flickr.checks




from collections import OrderedDict
from itertools import chain
from operator import attrgetter
import datetime

from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import InvalidPage
from django.urls import reverse
from django.db import models
from django.http import Http404
from django.shortcuts import redirect
from django.utils import six, timezone
from django.utils.encoding import force_str, force_text
from django.utils.translation import ugettext as _
from django.views.generic import DetailView, ListView, TemplateView, View
from django.views.generic import DayArchiveView as DjangoDayArchiveView

from .apps import ditto_apps
from .paginator import DiggPaginator

if ditto_apps.is_installed('flickr'):
    from ..flickr.models import Photo

if ditto_apps.is_installed('lastfm'):
    from ..lastfm.models import Scrobble

if ditto_apps.is_installed('pinboard'):
    from ..pinboard.models import Bookmark

if ditto_apps.is_installed('twitter'):
    from ..twitter.models import Tweet, User as TwitterUser


class PaginatedListView(ListView):
    """Use this instead of ListView to provide standardised pagination."""
    paginator_class = DiggPaginator
    paginate_by = 50
    page_kwarg = 'p'

    # See ditto.core.paginator for what these mean:
    paginator_body = 5 
    paginator_margin = 2 
    paginator_padding = 2
    paginator_tail = 2

    def __init__(self, **kwargs):
        return super().__init__(**kwargs)

    def paginate_queryset(self, queryset, page_size):
        """
        Paginate the queryset, if needed.

        This is EXACTLY the same as the standard ListView.paginate_queryset()
        except for this line:
            page = paginator.page(page_number, softlimit=True)
        Because we want to use the DiggPaginator's softlimit option.
        So that if you're viewing a page of, say, Flickr photos, and you switch
        from viewing by Uploaded Time to viewing by Taken Time, the new
        ordering might have fewer pages. In that case we want to see the final
        page, not a 404. The softlimit does that, but I can't see how to use
        it without copying all of this...
        """
        paginator = self.get_paginator(
            queryset,
            page_size,
            orphans = self.get_paginate_orphans(),
            allow_empty_first_page = self.get_allow_empty(),
            body    = self.paginator_body,
            margin  = self.paginator_margin,
            padding = self.paginator_padding,
            tail    = self.paginator_tail,
        )
        page_kwarg = self.page_kwarg
        page = self.kwargs.get(page_kwarg) or self.request.GET.get(page_kwarg) or 1
        try:
            page_number = int(page)
        except ValueError:
            if page == 'last':
                page_number = paginator.num_pages
            else:
                raise Http404(_("Page is not 'last', nor can it be converted to an int."))
        try:
            page = paginator.page(page_number, softlimit=True)
            return (paginator, page, page.object_list, page.has_other_pages())
        except InvalidPage as e:
            raise Http404(_('Invalid page (%(page_number)s): %(message)s') % {
                'page_number': page_number,
                'message': str(e)
            })


class DittoAppsMixin:
    """Contains methods for getting querysets for all the enabled Ditto apps.

    Contains a structure about all the apps and their 'varieties' (eg,
    'tweets' or 'favorites').

    Provides for 'pages' that will show one app+variety per page, based on a
    URL with app_slug and variety_slug keywords.

    NOTE: we have both *_name and *_slug to allow for us to change URLs. Which
    is the case, for example, with using /twitter/likes/ which has a
    variety_name of 'favorite' (rather than 'likes').
    """

    # These all set in set_app_and_variety():
    app_name = None  # eg, 'flickr', or 'twitter'.
    app_slug = None  # eg, 'flickr', or 'twitter'.
    variety_name = None  # eg, 'tweet' or 'favorite'.
    variety_slug = None  # eg, 'tweets' or 'likes'.

    # Set in __init__():
    apps = None


    def __init__(self, *args, **kwargs):

        self.apps = []

        enabled_apps = ditto_apps.enabled()

        # The order is important - the first app, and its first variety,
        # will be the defaults.

        if 'flickr' in enabled_apps:
            self.apps.append({
                'slug': 'flickr',
                'name': 'flickr',
                'varieties': [
                    {
                        'slug': 'photos',
                        'name': 'photo-uploaded',
                        'context_object_name': 'flickr_photo_list',
                        'queryset': Photo.public_objects.all().prefetch_related('user'),
                    },
                    {
                        # A bit cheeky as a slug, but seems to work:
                        'slug': 'photos/taken',
                        'name': 'photo-taken',
                        'date_field': 'taken_time',
                        'context_object_name': 'flickr_photo_list',
                        'queryset': Photo.public_objects.all().prefetch_related('user'),
                    },
                ],
            })

        if 'lastfm' in enabled_apps:
            self.apps.append({
                'slug': 'lastfm',
                'name': 'lastfm',
                'varieties': [
                    {
                        'slug': 'listens',
                        'name': 'scrobble',
                        'context_object_name': 'lastfm_scrobble_list',
                        'queryset': Scrobble.objects.all().prefetch_related('artist', 'track'),
                    },
                ],
            })

        if 'pinboard' in enabled_apps:
            self.apps.append({
                'slug': 'pinboard',
                'name': 'pinboard',
                'varieties': [
                    {
                        'slug': 'bookmarks',
                        'name': 'bookmark',
                        'context_object_name': 'pinboard_bookmark_list',
                        'queryset': Bookmark.public_objects.all().prefetch_related('account'),
                    },
                ],
            })

        if 'twitter' in enabled_apps:
            self.apps.append({
                'slug': 'twitter',
                'name': 'twitter',
                'varieties': [
                    {
                        'slug': 'tweets',
                        'name': 'tweet',
                        'context_object_name': 'twitter_tweet_list',
                        'queryset': Tweet.public_tweet_objects.all().prefetch_related('user'),
                    },
                    {
                        'slug': 'likes',
                        'name': 'favorite',
                        'context_object_name': 'twitter_favorite_list',
                        'queryset': Tweet.public_favorite_objects.all().prefetch_related('user'),
                    },
                ],
            })

        super().__init__(*args, **kwargs)

    def set_app_and_variety(self, **kwargs):
        """
        MUST be called by childs' get() methods, before they do much else.
        I don't like this reliance, but not sure how else to set app_name and
        variety_name before doing other get() stuff.

        Child get() methods should handle what happens if one or both of
        self.app_name and self.variety_name are left as None after this.

        Based on URL kwargs, sets the values of:
            * self.app_name
            * self.app_slug
            * self.variety_name
            * self.variety_slug
        So long as the 'app' and 'variety' slugs are valid.
        """
        app_slug = kwargs.get('app', None)
        variety_slug = kwargs.get('variety', None)

        if self.is_valid_app_slug(app_slug):
            self.app_name = self.get_app_name_from_slug(app_slug)

            if self.is_valid_variety_slug(app_slug, variety_slug):
                self.variety_name = self.get_variety_name_from_slugs(
                                                        app_slug, variety_slug)
            elif variety_slug:
                raise Http404(
                    "'%s' is not a valid variety slug for the '%s' app slug."%\
                                                    (variety_slug, app_slug))
        elif app_slug:
            raise Http404("'%s' is not a valid app slug."% app_slug)

        self.app_slug = app_slug
        self.variety_slug = variety_slug

    def get_date_field(self):
        return self.get_date_field_for_app_variety(
                                            self.app_name, self.variety_name)

    def get_context_object_name(self, object_list):
        """
        Get the name of the item to be used in the context.
        eg, 'twitter_favorite_list'.

        Overriding MultipleObjectMixin.get_context_object_name()
        """
        return self.get_context_object_name_for_app_variety(
                                            self.app_name, self.variety_name)

    def get_queryset(self):
        queryset = self.get_queryset_for_app_variety(
                                            self.app_name, self.variety_name)

        if queryset is None:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. "
                "%(cls)s.get_queryset() can't find the correct "
                "queryset for this app and variety."% {
                    'cls': self.__class__.__name__
                }
            )

        return queryset

    def get_context_data(self, **kwargs):
        """Add context stuff related to apps, varieties, etc."""
        context = super().get_context_data(**kwargs)
        context['ditto_app_name'] = self.app_name
        context['ditto_app_slug'] = self.app_slug
        context['ditto_variety_name'] = self.variety_name
        context['ditto_variety_slug'] = self.variety_slug
        return context


    # Below -- convenience methods for getting bits of data from the self.apps
    # structure.

    def get_app_varieties(self):
        """A list of tuples mapping app_name to each variety_name.
        May have duplicate app_names if the app has more than one variety."""
        app_varieties = []
        for app in self.apps:
            for variety in app['varieties']:
                app_varieties.append( (app['name'], variety['name']) )
        return app_varieties

    def get_app_slugs(self):
        """eg ['flickr', 'pinboard', 'twitter']."""
        return [app['slug'] for app in self.apps]

    def get_variety_slugs(self, app_slug):
        "A list of the slugs for the varieties of the app indicated by app_slug"
        app = self.get_app_from_slug(app_slug)
        return [ variety['slug'] for variety in app['varieties'] ]

    def is_valid_app_slug(self, app_slug):
        "Does this app slug exist in self.apps?"
        return (app_slug in self.get_app_slugs())

    def is_valid_variety_slug(self, app_slug, variety_slug):
        "Does this variety slug exist for app_slug's app in self.apps?"
        return (variety_slug in self.get_variety_slugs(app_slug))

    def get_default_app_slug(self):
        "Just the slug of the first app in the list."
        return self.apps[0]['slug']

    def get_default_variety_slug(self):
        "Just the slug of the first variety of the first app in the list."
        return self.apps[0]['varieties'][0]['slug']

    def get_default_variety_slug_for_app_slug(self, app_slug):
        "The slug of the first variety for the app indicated by app_slug."
        return self.get_app_from_slug(app_slug)['varieties'][0]['slug']

    def get_app_name_from_slug(self, app_slug):
        "What's the name of the app that has the slug app_slug?"
        return self.get_app_from_slug(app_slug)['name']

    def get_variety_name_from_slugs(self, app_slug, variety_slug):
        """What's the name of the variety that has the slug variety_slug, in
        the app that has the slug app_slug?"""
        variety = self.get_variety_from_slugs(app_slug, variety_slug)
        return variety['name']

    def get_app_from_slug(self, app_slug):
        "Get all the data in self.apps for the app with the slug app_slug"
        return list(filter(lambda a: a['slug'] == app_slug, self.apps))[0]

    def get_variety_from_slugs(self, app_slug, variety_slug):
        """Given app_slug and variety_slug, return the data in self.apps for
        that variety."""
        app = self.get_app_from_slug(app_slug)
        varieties = app['varieties']
        return list(filter(lambda v: v['slug'] == variety_slug, varieties))[0]

    def get_app_from_name(self, app_name):
        "Get all the data in self.apps for the app with the name app_name"
        return list(filter(lambda a: a['name'] == app_name, self.apps))[0]

    def get_variety_from_names(self, app_name, variety_name):
        """Given app_name and variety_name, return the data in self.apps for
        that variety."""
        app = self.get_app_from_name(app_name)
        varieties = app['varieties']
        return list(filter(lambda v: v['name'] == variety_name, varieties))[0]

    def get_app_slug_from_name(self, app_name):
        return self.get_app_from_name(app_name)['slug']

    def get_variety_slug_from_names(self, app_name, variety_name):
        return self.get_variety_from_names(app_name, variety_name)['slug']

    def get_context_object_name_for_app_variety(self, app_name, variety_name):
        """Given app_name and variety_name, return the context_object_name in
        self.apps for that variety."""
        variety = self.get_variety_from_names(app_name, variety_name)
        return variety['context_object_name']

    def get_queryset_for_app_variety(self, app_name, variety_name):
        """Given app_name and variety_name, return the queryset in
        self.apps for that variety."""
        variety = self.get_variety_from_names(app_name, variety_name)
        return variety['queryset']

    def get_date_field_for_app_variety(self, app_name, variety_name):
        variety = self.get_variety_from_names(app_name, variety_name)
        try:
            return variety['date_field']
        except KeyError:
            return 'post_time'


class HomeView(DittoAppsMixin, TemplateView):
    template_name = 'ditto/home.html'

    # How many things do we list on the page?
    items_to_list = 30

    # What we want to display:
    app_varieties_to_display = [
        ('flickr', 'photo-uploaded'),
        ('pinboard', 'bookmark'),
        ('twitter', 'tweet'),

        # Things we could display, but are currently hidden:

        # Otherwise we'll get two sets of photos, sorted by upload or taken:
        #('flickr', 'photo-taken'),

        # No easy way to tell in the template whether a tweet is posted
        # or favorited, so remove the favorites:
        #('twitter', 'favorite'),

        # These are liable to swamp out all other things:
        #('lastfm', 'scrobble'),
    ]

    # Hacky.
    # Set to True to include Tweets that are replies:
    include_twitter_replies = False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Put all querysets for available apps in this list:
        querysets = []

        for app_name, variety_name in self.get_app_varieties_to_display():
            qs = self.get_queryset_for_app_variety(app_name, variety_name)

            if self.include_twitter_replies == False:
                # Don't want to include Tweets that are replies on the home page.
                if app_name == 'twitter' and variety_name == 'tweet':
                    qs = qs.filter(in_reply_to_screen_name__exact='')

            querysets.append( qs[:self.items_to_list] )

        # Aggregate the individual items from all querysets into one list,
        # sorted by post_time descending.
        context['object_list'] = sorted(
            chain(*querysets),
            key=attrgetter('post_time'),
            reverse=True
        )[:self.items_to_list]

        return context

    def get_app_varieties_to_display(self):
        """
        Get the union of self.app_varieties and the varieties that are
        actually installed.
        """
        available_app_varieties = self.get_app_varieties()
        to_display = []

        for app_variety in self.app_varieties_to_display:
            if app_variety in available_app_varieties:
                to_display.append(app_variety)

        return to_display


class DayArchiveView(DittoAppsMixin, DjangoDayArchiveView):
    """A single app+variety for a single day.
    eg /2016/04/20/twitter/likes
    """
    template_name = 'ditto/archive_day.html'
    month_format = '%m'
    allow_empty = True
    allow_future = False

    # All DittoItem-inheriting classes have 'post_time':
    #date_field = 'post_time'

    def get(self, request, *args, **kwargs):
        """Sets up self.app_name etc, and handles redirects to defaults if
        both are missing, or if app_name is valid but has no variety.
        """

        self.set_app_and_variety(**kwargs)

        if self.variety_name is None:
            if self.app_name is None:
                # Redirect to default app and variety for this date.
                return redirect(
                    reverse('ditto:day_archive', kwargs={
                        'year':     self.get_year(),
                        'month':    self.get_month(),
                        'day':      self.get_day(),
                        'app':      self.get_default_app_slug(),
                        'variety':  self.get_default_variety_slug(),
                }) )
            else:
                # Redirect to the default variety for this app.
                return redirect(
                    reverse('ditto:day_archive', kwargs={
                        'year': self.get_year(),
                        'month': self.get_month(),
                        'day': self.get_day(),
                        'app': self.app_slug,
                        'variety': self.get_default_variety_slug_for_app_slug(
                                                                self.app_slug),
                }) )

        return super().get(request, *args, **kwargs)


    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['variety_counts'] = self.get_variety_counts()
        # Useful to display photos-taken properly:
        context['date_field'] = self.get_date_field_for_app_variety(
                                            self.app_name, self.variety_name)
        return context

    def get_variety_counts(self):
        """
        For all of the querysets for all enabled apps, get the count of items
        they have for this day.

        Returns an OrderedDict like:
            [ ('twitter_tweet_list', 6), (...), ]

        Most of this is adapted from standard DayArchiveView methods, so that
        we can run it against each queryset in turn, rather than only
        self.queryset.

        And we need to set date_field for each variety in turn, rather than
        assume each variety uses self.date_field.
        """

        # From get_dated_items():
        year = self.get_year()
        month = self.get_month()
        day = self.get_day()
        date = _date_from_string(year, self.get_year_format(),
                                 month, self.get_month_format(),
                                 day, self.get_day_format())

        # From _make_single_date_lookup():
        since = self._make_date_lookup_arg(date)
        until = self._make_date_lookup_arg(date + datetime.timedelta(days=1))

        # Adapted from get_dated_queryset():

        allow_future = self.get_allow_future()
        allow_empty = self.get_allow_empty()

        # Want to keep them in the same order as get_queryset_names() provides.
        counts = []

        for app_name, variety_name in self.get_app_varieties():
            date_field = self.get_date_field_for_app_variety(
                                                        app_name, variety_name)
            qs = self.get_queryset_for_app_variety(app_name, variety_name)
            qs = qs.filter(**{
                '%s__gte' % date_field: since,
                '%s__lt' % date_field: until,
            })
            paginate_by = self.get_paginate_by(qs)
            #if not allow_future:
                #now = timezone.now() if self.uses_datetime_field else timezone_today()
                #qs = qs.filter(**{'%s__lte' % date_field: now})
            if not allow_empty:
                # When pagination is enabled, it's better to do a cheap query
                # than to load the unpaginated queryset in memory.
                is_empty = len(qs) == 0 if paginate_by is None else not qs.exists()
                if is_empty:
                    raise Http404(_("No %(verbose_name_plural)s available") % {
                        'verbose_name_plural': force_text(qs.model._meta.verbose_name_plural)
                    })

            counts.append({
                'count':        qs.count(),
                'app_name':     app_name,
                'variety_name': variety_name,
                'app_slug':     self.get_app_slug_from_name(app_name),
                'variety_slug': self.get_variety_slug_from_names(
                                                    app_name, variety_name),
                # Is this the current variety of the page?
                'is_active':    (app_name == self.app_name and variety_name == self.variety_name),
            })

        return counts


#class TagListView(TemplateView):
    #"Doesn't really do anything at the moment."
    #template_name = 'ditto/tag_list.html'


#class TagDetailView(TemplateView):
    #"All items with a certain tag"
    #template_name = 'ditto/tag_detail.html'

    #def get_context_data(self, **kwargs):
        #context = super().get_context_data(**kwargs)
        #context['tag'] = kwargs['slug']
        #if ditto_apps.is_enabled('pinboard'):
            #context['pinboard_bookmark_list'] = Bookmark.public_objects.filter(
                                            #tags__slug__in=[kwargs['slug']])
        #return context


def _date_from_string(year, year_format, month, month_format, day='', day_format='', delim='__'):
    """
    Helper: get a datetime.date object given a format string and a year,
    month, and possibly day; raise a 404 for an invalid date.

    From django.views.generic.dates
    """
    format = delim.join((year_format, month_format, day_format))
    datestr = delim.join((year, month, day))
    try:
        return datetime.datetime.strptime(force_str(datestr), format).date()
    except ValueError:
        raise Http404(_(u"Invalid date string '%(datestr)s' given format '%(format)s'") % {
            'datestr': datestr,
            'format': format,
        })





# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-10-12 10:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('lastfm', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scrobble',
            name='account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='scrobbles', to='lastfm.Account'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-16 10:36
from __future__ import unicode_literals

import ditto.flickr.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flickr', '0015_auto_20160514_1456'),
    ]

    operations = [
        migrations.AddField(
            model_name='photo',
            name='video_original_file',
            field=models.FileField(blank=True, default='', help_text='Only present for Videos.', upload_to=ditto.flickr.models.Photo.upload_path),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import taggit.managers


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('pinboard', '0012_account_is_active'),
    ]

    operations = [
        migrations.CreateModel(
            name='BookmarkTag',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('name', models.CharField(unique=True, verbose_name='Name', max_length=100)),
                ('slug', models.SlugField(unique=True, verbose_name='Slug', max_length=100)),
                ('time_created', models.DateTimeField(help_text='The time this item was created in the database.', auto_now_add=True)),
                ('time_modified', models.DateTimeField(help_text='The time this item was last saved to the database.', auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TaggedBookmark',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('object_id', models.IntegerField(verbose_name='Object id', db_index=True)),
                ('time_created', models.DateTimeField(help_text='The time this item was created in the database.', auto_now_add=True)),
                ('time_modified', models.DateTimeField(help_text='The time this item was last saved to the database.', auto_now=True)),
                ('content_type', models.ForeignKey(related_name='pinboard_taggedbookmark_tagged_items', to='contenttypes.ContentType', verbose_name='Content type', on_delete=models.CASCADE)),
                ('tag', models.ForeignKey(related_name='pinboard_taggedbookmark_items', to='pinboard.BookmarkTag', on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='bookmark',
            name='tags',
            field=taggit.managers.TaggableManager(verbose_name='Tags', through='pinboard.TaggedBookmark', to='pinboard.BookmarkTag', help_text='A comma-separated list of tags.'),
        ),
    ]



# coding: utf-8
import hashlib
from django.core.validators import URLValidator
from django.db import models

from taggit.managers import TaggableManager
from taggit.models import GenericTaggedItemBase, TagBase

from .managers import _BookmarkTaggableManager, PublicToreadManager, ToreadManager
from ..core.models import DittoItemModel, TimeStampedModelMixin


class Account(TimeStampedModelMixin, models.Model):
    username = models.CharField(null=False, blank=False, max_length=30,
                unique=True,
                help_text="eg, 'philgyford'")
    url = models.URLField(max_length=255, null=False, blank=False,
                unique=True,
                verbose_name='URL',
                help_text="eg, 'https://pinboard.in/u:philgyford'")
    # max_length derived from DittoAccount.username max_length plus
    # 21 characters for ':12345...'.
    api_token = models.CharField(null=False, blank=False, max_length=51,
                    verbose_name='API Token',
                    help_text='eg, "philgyford:1234567890ABCDEFGHIJ"')
    is_active = models.BooleanField(default=True, null=False, blank=False,
                        help_text="If false, new Bookmarks won't be fetched.")

    def __str__(self):
        return self.username

    class Meta:
        ordering = ['username']

    def get_absolute_url(self):
        from django.urls import reverse
        return reverse('pinboard:account_detail', kwargs={'username': self.username})

    @property
    def public_bookmarks_count(self):
        return Bookmark.public_objects.count()


class ExtraBookmarkManagers(models.Model):
    """Managers to use in the Bookmark model, in addition to the defaults
    defined in DittoItemModel.
    These need to be here, rather than in the Pinboard model, or they will
    override those in DittoItemModel.
    """
    toread_objects = ToreadManager()
    public_toread_objects = PublicToreadManager()

    class Meta:
        abstract = True


class BookmarkTag(TimeStampedModelMixin, TagBase):
    """Our custom version of a Taggit Tag model, for use with Bookmarks.

    NOTE: If you create two tags in Pinboard, with names "dog" and "DOG",
    they will both get the slug "dog". Taggit doesn't work like that by
    default, as both Name and Slug are unique. So the second of those
    tags that you create will have a slug of "dog_1". Not sure how best to fix
    that, if at all.
    """

    class Meta:
        verbose_name = "Tag"
        verbose_name_plural = "Tags"

    def slugify(self, tag, i=None):
        """Pinboard's slugs for tags don't encode many things. Most unicode
        characters are allowed as-is, and only a few special characters
        are URL encoded. This may not be an exhaustive list.

        We also lowercase the slugs. Which isn't exactly what Pinboard does,
        but it makes it much easier for us to show related Bookmarks, no matter
        if they're tagged with 'Fish' or 'fish'.

        This is inherited from TagBase.

        `tag` is a string, like 'mytag'
        `i` is either None or an integer, which signifies how many times the
            slug for this tag has been attempted to be calculated, it is None
            on the first time, and the counting begins at 1 thereafter.
        """
        tag = tag.lower()
        tag = tag.replace('%', '%25')
        replace = {
                '#':    '%23',
                '&':    '%2526',
                "'":    '%27',
                '+':    '%252B',
                '/':    '%252f',
                '?':    '%3f',
                '"':    '%22',
                '<':    '%3C',
                '>':    '%3E',
                '\\':   '%5c',
                # These shouldn't be in the tag name, but just in case:
                ',':    '',
                ' ':    '-',
        }

        for f, r in replace.items():
            tag = tag.replace(f, r)

        # So if you slugify "dog" for one tag and then "DOG" for another,
        # the first will have a slug of "dog" and the second will have "dog_1".

        if i is not None:
            tag += "_%d" % i

        return tag


class TaggedBookmark(TimeStampedModelMixin, GenericTaggedItemBase):
    tag = models.ForeignKey(BookmarkTag, on_delete=models.CASCADE,
                            related_name="%(app_label)s_%(class)s_items")


class Bookmark(DittoItemModel, ExtraBookmarkManagers):

    ditto_item_name = 'pinboard_bookmark'

    # Properties inherited from DittoItemModel:
    #
    # title         (CharField)
    # permalink     (URLField)
    # summary       (CharField)
    # is_private    (BooleanField)
    # fetch_time    (DateTimeField, UTC)
    # post_time     (DateTimeField, UTC)
    # latitude      (DecimalField)
    # longitude     (DecimalField)
    # raw           (TextField)

    account = models.ForeignKey(Account, on_delete=models.CASCADE,
                                                    null=False, blank=False)

    # `url` in the Pinboard API:
    url = models.TextField(null=False, blank=False,
                                                validators=[URLValidator()])

    # `extended` in the Pinboard API:
    description = models.TextField(null=False, blank=True,
                    help_text="The 'extended' text description.")

    # `toread` in the Pinboard API:
    to_read = models.BooleanField(default=False, null=False, blank=False)

    # 'shared' in the Pinboard API is inverted and saved as
    # DittoItem::is_private.

    # Pinboard usese some kind of hash for each individual URL, but these
    # don't come back via the API. So we'll make our own (in self.save()).
    url_hash = models.CharField(
        null=False, blank=True, max_length=12, db_index=True,
                                help_text="Slug in the Bookmark's local URL.")

    # Up to 100 tags
    # Up to 255 chars each. No commas or whitespace.
    # Private tags start with a period.
    tags = TaggableManager(manager=_BookmarkTaggableManager,
                                                        through=TaggedBookmark)

    class Meta:
        ordering = ['-post_time']
        unique_together = (('account', 'url'),)

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        """Create a url_hash for this bookmark if it doesn't have one.
        This is not the best ever way to do this. For one, there's a very slim
        chance of clashes. But it seems good enough for now.
        """
        if not self.url_hash:
            self.url_hash = hashlib.md5(
                                    self.url.encode('utf-8')).hexdigest()[:12]
        super().save(*args, **kwargs)

    def get_absolute_url(self):
        from django.urls import reverse
        return reverse('pinboard:bookmark_detail',
            kwargs={'username': self.account.username, 'hash': self.url_hash})

    def get_next_public_by_post_time(self):
        "The next public Bookmark by this Account, ordered by post_time."
        try:
            return Bookmark.public_objects.filter(post_time__gte=self.post_time, account=self.account).exclude(pk=self.pk).order_by('post_time')[:1].get()
        except:
            pass

    def get_previous_public_by_post_time(self):
        "The previous public Bookmark by this Account, ordered by post_time."
        try:
            return Bookmark.public_objects.filter(post_time__lte=self.post_time, account=self.account).exclude(pk=self.pk).order_by('-post_time')[:1].get()
        except:
            pass

    # Shortcuts:
    def get_next(self):
        return self.get_next_public_by_post_time()

    def get_previous(self):
        return self.get_previous_public_by_post_time()

    def slugs_match_tags(self, slugs):
        """Does a list of slugs equal the slugs of all this Bookmark's tags?
        Keyword arguments:
        slugs -- A list of slugs, eg ['carrot', 'pea']
        Returns boolean
        """
        return set(self.tags.slugs()) == set(slugs)

    def _summary_source(self):
        "Used to make the `summary` property."
        return self.description




import factory

from . import models
from ..core.utils import datetime_now


class AccountFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Account

    username = factory.Sequence(lambda n: 'user%04d' % n)
    realname = factory.Sequence(lambda n: 'Real Name %04d' % n)
    api_key = factory.Sequence(lambda n: 'api-key-%04d' % n)


class ArtistFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Artist

    name = factory.Sequence(lambda n: 'Artist %04d' % n)
    slug = factory.Sequence(lambda n: 'artist+%04d' % n)
    original_slug = factory.Sequence(lambda n: 'Artist+%04d' % n)


class TrackFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Track

    name = factory.Sequence(lambda n: 'Track %04d' % n)
    slug = factory.Sequence(lambda n: 'track+%04d' % n)
    original_slug = factory.Sequence(lambda n: 'Track+%04d' % n)
    artist = factory.SubFactory(ArtistFactory)


class AlbumFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Album

    name = factory.Sequence(lambda n: 'Album %04d' % n)
    slug = factory.Sequence(lambda n: 'album+%04d' % n)
    original_slug = factory.Sequence(lambda n: 'Album+%04d' % n)
    artist = factory.SubFactory(ArtistFactory)


class ScrobbleFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Scrobble

    account = factory.SubFactory(AccountFactory)
    track = factory.SubFactory(TrackFactory)
    artist = factory.SubFactory(ArtistFactory)
    post_time = datetime_now()




# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-11-21 13:24
from __future__ import unicode_literals

from django.db import migrations, models


def set_post_year(apps, schema_editor):
    """
    Sets the `post_year` on every Bookmark.
    """
    Bookmark = apps.get_model('pinboard', 'Bookmark')
    for row in Bookmark.objects.all():
        row.post_year = row.post_time.year
        row.save()


class Migration(migrations.Migration):

    dependencies = [
        ('pinboard', '0023_auto_20161117_1622'),
    ]

    operations = [
        migrations.AddField(
            model_name='bookmark',
            name='post_year',
            field=models.PositiveSmallIntegerField(blank=True, db_index=True, help_text='Set automatically on save', null=True),
        ),

        migrations.RunPython(set_post_year, reverse_code=migrations.RunPython.noop),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0031_auto_20160413_0906'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tweet',
            name='latitude',
            field=models.DecimalField(decimal_places=6, blank=True, null=True, max_digits=9),
        ),
        migrations.AlterField(
            model_name='tweet',
            name='longitude',
            field=models.DecimalField(decimal_places=6, blank=True, null=True, max_digits=9),
        ),
    ]



from django.conf import settings


# Creating all the defaults for settings.
# In our code, if we want to use a DITTO_TWITTER_* setting we should import
# from here, not django.conf.settings.

DITTO_TWITTER_DIR_BASE = getattr(settings, 'DITTO_TWITTER_DIR_BASE', 'twitter')

DITTO_TWITTER_USE_LOCAL_MEDIA = getattr(settings,
                                    'DITTO_TWITTER_USE_LOCAL_MEDIA', False)





from . import FetchPhotosCommand
from ...fetch.multifetchers import RecentPhotosMultiAccountFetcher


class Command(FetchPhotosCommand):
    """Fetches photos from Flickr.

    For all accounts:
        ./manage.py fetch_flickr_photos --days=3
        ./manage.py fetch_flickr_photos --days=all

    For one account:
        ./manage.py fetch_flickr_photos --account=35034346050@N01 --days=3
        ./manage.py fetch_flickr_photos --account=35034346050@N01 --days=all

    """

    help = "Fetches recent or all photos for one or all Flickr Accounts"

    days_help = 'Fetches the most recent or all Photos, eg "3" or "all".'

    def fetch_photos(self, nsid, days):
        return RecentPhotosMultiAccountFetcher(nsid=nsid).fetch(days=days)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pinboard', '0014_auto_20151028_1556'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bookmark',
            name='post_time',
            field=models.DateTimeField(help_text='The time the item was originally posted/created on its service.', blank=True, null=True),
        ),
    ]



import datetime
import pytz
import factory

from . import models
from ..core.utils import datetime_now


class UserFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.User

    twitter_id = factory.Sequence(lambda n: (n * 10000))
    screen_name = factory.Sequence(lambda n: 'user%d' % n)
    name = factory.Sequence(lambda n: 'User Name %d' % n)

    created_at = factory.LazyAttribute(lambda o:
                               datetime_now() - datetime.timedelta(weeks=52)
                            )

    description = factory.Sequence(lambda n: 'A description %d' % n)
    is_private = False
    fetch_time = datetime_now()

    avatar = factory.django.ImageField(filename='my_avatar.jpg')


class AccountFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Account

    user = factory.SubFactory(UserFactory,
            twitter_id=factory.Sequence(lambda n: n),
            screen_name=factory.Sequence(lambda n: 'user%d' % n)
        )

class AccountWithCredentialsFactory(AccountFactory):
    """We only want to add these when we're going to be testing the
    fetching of API data for the Account's user.
    """
    consumer_key = 'TESTCONSUMERKEY'
    consumer_secret = 'TESTCONSUMERSECRET'
    access_token = factory.Sequence(lambda n: '%d-TESTACCESSTOKEN' % n)
    access_token_secret = 'TESTACCESSTOKENSECRET'


class TweetFactory(factory.DjangoModelFactory):

    class Meta:
        model = models.Tweet

    user = factory.SubFactory(UserFactory)
    text = factory.Sequence(lambda n: 'The text of tweet %d' % n)
    twitter_id = factory.Sequence(lambda n: (n * 10000000))
    fetch_time = datetime_now()

    post_time = factory.LazyAttribute(lambda o:
                                datetime_now() - datetime.timedelta(weeks=4)
                            )
    source = 'web'


class MediaFactory(factory.DjangoModelFactory):
    "Parent class for the photo, video and gif factories."

    class Meta:
        model = models.Media

    twitter_id = factory.Sequence(lambda n: (n * 10000000))
    large_w = 640
    large_h = 360
    medium_w = 600
    medium_h = 338
    small_w = 340
    small_h = 191
    thumb_w = 150
    thumb_h = 150

    image_file = factory.django.ImageField(filename='example.jpg')
    mp4_file = factory.django.FileField(filename='example.mp4')

    @factory.post_generation
    def tweets(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for tweet in extracted:
                self.tweets.add(tweet)


class PhotoFactory(MediaFactory):

    media_type = 'photo'

    large_w = 938
    large_h = 397
    medium_w = 600
    medium_h = 253
    small_w = 340
    small_h = 143
    thumb_w = 150
    thumb_h = 150

    image_url = factory.Sequence(
                            lambda n: 'http://pbs.twimg.com/media/%d.jpg' % n)
    image_file = factory.django.ImageField(filename='example.jpg')


class VideoFactory(MediaFactory):

    media_type = 'video'

    image_url = factory.Sequence(
        lambda n: 'http://pbs.twimg.com/ext_tw_video_thumb/%d/pu/img/%d.jpg' % (n, n))

    xmpeg_url = factory.Sequence(lambda n: 'https://video.twimg.com/ext_tw_video/%d/pu/pl/%d.m3u8' % (n, n))
    dash_url = factory.Sequence(lambda n: 'https://video.twimg.com/ext_tw_video/%d/pu/pl/%d.mpd' % (n, n))

    aspect_ratio = '16:9'
    duration = 20000

    image_file = factory.django.ImageField(filename='example.jpg')


class AnimatedGifFactory(MediaFactory):

    media_type = 'animated_gif'

    image_url = factory.Sequence(
        lambda n: 'http://pbs.twimg.com/ext_tw_video_thumb/%d/pu/img/%d.jpg' % (n, n))

    mp4_url = factory.Sequence(lambda n: 'https://pbs.twimg.com/tweet_video/%d.mp4' % n)

    aspect_ratio = '16:9'

    image_file = factory.django.ImageField(filename='example.jpg')
    mp4_file = factory.django.FileField(filename='example.mp4')




from django.db import models


class PublicItemManager(models.Manager):
    """
    Only returns items that are public.
    Should be used on ALL public pages of the site.
    """
    def get_queryset(self):
        return super().get_queryset().filter(is_private=False)




# coding: utf-8
import datetime
import json
import os
import pytz

from .fetch.savers import TweetSaver
from .models import User
from ..core.utils import datetime_now


class IngestError(Exception):
    pass


class TweetIngester(object):
    """For importing a downloaded archive of tweets.
    Request yours from https://twitter.com/settings/account

    Use like:
        results = TweetIngester.ingest('/Users/phil/Downloads/12552_dbeb4be9b8ff5f76d7d486c005cc21c9faa61f66/data/js/tweets')

    Where that's the path to the directory containing the *.js files holding
    tweet data.

    results will be a dict of data about what happened, including
    results['success'] which is boolean.
    """

    def __init__(self):
        # Used as the 'fetch_time' for each tweet.
        self.fetch_time = datetime_now()

        # How many .js files we loaded the tweets from:
        self.file_count = 0

        # How mnay tweets we found in all the files:
        self.tweet_count = 0

        # Stores all the imported data from the files before saving.
        # So that we know we've got through all the files within JSON errors
        # etc before we begin touching the DB.
        self.tweets_data = []

    def ingest(self, directory):
        """Import all the tweet data and create/update the tweets."""

        self._load_data(directory)
        self._save_tweets()
        if self.tweet_count > 0:
            return {'success': True,
                    'tweets': self.tweet_count,
                    'files': self.file_count, }
        else:
            return {'success': False,
                    'tweets': 0,
                    'files': self.file_count,
                    'messages': ["No tweets were found"], }

    def _load_data(self, directory):
        """Goes through all the *.js files in `directory` and puts the tweet
        data inside into self.tweets_data.

        No data is saved to the database until we've successfully loaded JSON
        from all of the files.

        Keyword arguments:
        directory -- The directory to load the files from.

        Raises:
        FetchError -- If the directory is invalid, or there are no .js files,
            or we can't load JSON from one of the files.
        """
        try:
            for file in os.listdir(directory):
                if file.endswith('.js'):
                    filepath = '%s/%s' % (directory, file)
                    self._get_data_from_file(filepath)
                    self.file_count += 1
        except OSError as e:
            raise IngestError(e)

        if self.file_count == 0:
            raise IngestError("No .js files found in %s" % directory)

    def _get_data_from_file(self, filepath):
        """Looks in a file, parses its JSON, and adds a dict of data about
        each tweet found to self.tweets_data.

        Arguments:
        filespath -- Absolute path to the file.
        """
        f = open(filepath, 'r')
        lines = f.readlines()
        # Remove first line, that contains JavaScript:
        lines = lines[1:]
        try:
            tweets_data = json.loads(''.join(lines))
        except ValueError:
            raise IngestError("Could not load JSON from %s" % filepath)
        else:
            self.tweets_data.extend(tweets_data)
        f.close()

    def _save_tweets(self):
        """Go through the list of dicts that is self.tweets_data and
        create/update each tweet in the DB.
        """
        if len(self.tweets_data) == 0:
            return

        for tweet in self.tweets_data:
            TweetSaver().save_tweet(tweet, self.fetch_time)
            self.tweet_count += 1




# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-10-26 15:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lastfm', '0002_auto_20161012_1012'),
    ]

    operations = [
        migrations.AddField(
            model_name='album',
            name='original_slug',
            field=models.TextField(blank=True, help_text='As used on Last.fm. Mixed case.', null=True),
        ),
        migrations.AddField(
            model_name='artist',
            name='original_slug',
            field=models.TextField(blank=True, help_text='As used on Last.fm. Mixed case.', null=True),
        ),
        migrations.AddField(
            model_name='track',
            name='original_slug',
            field=models.TextField(blank=True, help_text='As used on Last.fm. Mixed case.', null=True),
        ),
        migrations.AlterField(
            model_name='album',
            name='slug',
            field=models.TextField(db_index=True, help_text='Lowercase'),
        ),
        migrations.AlterField(
            model_name='artist',
            name='slug',
            field=models.TextField(help_text='Lowercase', unique=True),
        ),
        migrations.AlterField(
            model_name='track',
            name='slug',
            field=models.TextField(db_index=True, help_text='Lowercase'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0027_remove_user_profile_image_url'),
    ]

    operations = [
        migrations.RenameField(
            model_name='user',
            old_name='favorites_count',
            new_name='favourites_count',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-06-13 16:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('twitter', '0043_remove_media_is_private'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tweet',
            name='text',
            field=models.TextField(),
        ),
    ]



from django.conf import settings
from django.core.checks import register, Error, Tags


@register(Tags.compatibility)
def check_taggit_is_installed(app_configs=None, **kwargs):
    checks = []
    try:
        from taggit import models

    except ImportError:
        checks.append(
            Error(
                "The django-taggit app is required to use ditto.flickr.",
                hint=("Install django-taggit"),
                id='ditto.flickr.E001',
            )
        )

    if len(checks) == 0:
        if 'taggit' not in settings.INSTALLED_APPS:
            checks.append(
                Error(
                    "The django-taggit app must be in INSTALLED_APPS",
                    hint=("Add 'taggit' to INSTALLED_APPS "
                            "in your settings file."),
                    id='ditto.flickr.E002',
                )
            )

    return checks





# coding: utf-8
from django.db import models
from django.forms.models import model_to_dict

from .managers import PublicItemManager
from .utils import truncate_string


class TimeStampedModelMixin(models.Model):
    "Should be mixed in to all models."
    time_created = models.DateTimeField(auto_now_add=True,
                help_text="The time this item was created in the database.")
    time_modified = models.DateTimeField(auto_now=True,
                help_text="The time this item was last saved to the database.")

    class Meta:
        abstract = True


class DiffModelMixin(object):
    """A model mixin that tracks model fields' values and provide some useful
    api to know what fields have been changed.

    eg:
    Get an object from the database.
    Set some of its properties.
    Call `myObj.has_changed` to see if any fields are different to in the DB.

    From http://stackoverflow.com/a/13842223/250962
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initial = self._dict

    @property
    def diff(self):
        """Returns a dict of properties that have changed, with values a list
        of before/after changes. eg:
        `{'categories': (None, [1, 3, 5]), 'rank': (0, 42)}`
        """
        d1 = self.__initial
        d2 = self._dict
        diffs = [(k, (v, d2[k])) for k, v in d1.items() if v != d2[k]]
        return dict(diffs)

    @property
    def has_changed(self):
        "Returns True if any of the properties have changed."
        return bool(self.diff)

    @property
    def changed_fields(self):
        "Returns a list of property names for any that have changed."
        return self.diff.keys()

    def get_field_diff(self, field_name):
        "Returns a diff for field if it's changed and None otherwise."
        return self.diff.get(field_name, None)

    def save(self, *args, **kwargs):
        "Saves model and set initial state."
        super().save(*args, **kwargs)
        self.__initial = self._dict

    @property
    def _dict(self):
        return model_to_dict(self, fields=[field.name for field in
                             self._meta.fields])


class DittoItemModel(TimeStampedModelMixin, DiffModelMixin, models.Model):
    """
    A content item on whatever service we're copying.
    eg, a Tweet, a Photo, a Bookmark, etc.

    Should be inherited by a model in child apps.
    eg, Tweet, Photo, Bookmark, etc.
    """

    # Should be overridden for child classes.
    # eg, 'flickr_photo', 'twitter_tweet', etc.
    # Used in templates.
    ditto_item_name = 'set__ditto_item_name__in_child_class'

    title = models.CharField(blank=True, max_length=255)
    permalink = models.URLField(blank=True,
                    help_text="URL of the item on the service's website.")
    # Ensures that all children have a common short piece of text for display:
    summary = models.CharField(blank=True, max_length=255,
        help_text="eg, Brief summary or excerpt of item's text content. No linebreaks or HTML.")
    is_private = models.BooleanField(default=False,
        help_text="If true, this item will not be shown on public-facing pages.")
    fetch_time = models.DateTimeField(null=True, blank=True,
                        help_text="The time the item's data was last fetched.")

    post_time = models.DateTimeField(null=True, blank=True, db_index=True,
        help_text="The time the item was originally posted/created on its service.")
    post_year = models.PositiveSmallIntegerField(
                                    null=True, blank=True, db_index=True,
                                    help_text="Set automatically on save")

    # Obviously not relevant to some items, like Bookmarks.
    latitude = models.DecimalField(null=True, blank=True,
                                            max_digits=9, decimal_places=6)
    longitude = models.DecimalField(null=True, blank=True,
                                            max_digits=9, decimal_places=6)

    raw = models.TextField(blank=True,
                                    help_text="eg, the raw JSON from the API.")

    # All Items (eg, used in Admin):
    objects = models.Manager()

    # All Items which aren't private. Should ALWAYS be used for public pages:
    public_objects = PublicItemManager()

    class Meta:
        abstract = True
        get_latest_by = 'post_time'

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        self.summary = self._make_summary()
        if self.post_time:
            self.post_year = self.post_time.year
        else:
            self.post_year = None
        super().save(*args, **kwargs)

    def _summary_source(self):
        """
        Child classes can return a string that's used to make the truncated,
        HTML-free `summary`.
        e.g.:
            return self.description
        """
        return ''

    def _make_summary(self):
        """
        Returns the string to be used for the `summary` property.
        """
        return truncate_string(self._summary_source(),
                                strip_html=True,
                                chars=255,
                                truncate='…',
                                at_word_boundary=True).replace('\n', ' ')\
                                                      .replace('\r', ' ')




from . import UpdateTwitterCommand
from ...fetch.fetchers import TweetsFetcher


class Command(UpdateTwitterCommand):
    """Fetches data for all Twitter Tweets in the DB, updating their info.

    Specify an account to use its API credentials:
    ./manage.py update_twitter_tweets --account=philgyford
    """

    help = "Fetches the latest data about each Twitter Tweet"

    singular_noun = 'Tweet'
    plural_noun = 'Tweets'

    def fetch(self, screen_name):
        return TweetsFetcher(screen_name=screen_name).fetch()




from datetime import datetime

from django.db import models


class WithScrobbleCountsManager(models.Manager):
    """
    Adds a with_scrobble_counts() method.
    """

    # Can we filter these (things) by Album?
    is_filterable_by_album = True

    # Can we filter these (things) by Artist?
    is_filterable_by_artist = True

    # Can we filter these (things) by Track?
    is_filterable_by_track = True

    def with_scrobble_counts(self, **kwargs):
        """
        Adds a `scrobble_count` field to the Queryset's objects, and
        orders the results by that, descending.

        eg:
            # All Tracks, each with a scrobble_count:
            Track.objects.with_scrobble_counts()

            # All Albums, each with a total scrobble_count:
            Album.objects.with_scrobble_counts()

            # All Artists, each with a total scrobble_count:
            Artist.objects.with_scrobble_counts()

            # Tracks by artist_obj:
            Track.objects.with_scrobble_counts(artist=artist_obj)

            # Tracks appearing on album_obj:
            Track.objects.with_scrobble_counts(album=album_obj)

            # Albums on which track_obj appears:
            Album.objects.with_scrobble_counts(track=track_obj)

        or combine filters:

            # Tracks by artist_obj, scrobbled by account_obj between
            # datetime_obj_1 and datetime_obj2:
            Track.objects.with_scrobble_counts(
                account         = account_obj,
                artist          = artist_obj,
                min_post_time   = datetime_obj_1,
                max_post_time   = datetime_obj_2,
            )

        Include an `account` to only include Scrobbles by that Account.
        Include an `album` to only include Scrobbles on that Album.
        Include an `artist` to only include Scrobbles by that Artist.
        Include a `track` to only include Scrobbles including that Track.
        Include a `min_post_time` to only include Scrobbles after then.
        Include a `max_post_time` to only include Scrobbles before then.
        """
        account         = kwargs.get('account', None)
        min_post_time   = kwargs.get('min_post_time', None)
        max_post_time   = kwargs.get('max_post_time', None)
        album           = kwargs.get('album', None)
        artist          = kwargs.get('artist', None)
        track           = kwargs.get('track', None)

        if album and not self.is_filterable_by_album:
            raise ValueError('This is not filterable by album')

        if artist and not self.is_filterable_by_artist:
            raise ValueError('This is not filterable by artist')

        if track and not self.is_filterable_by_track:
            raise ValueError('This is not filterable by track')

        if account is not None and account.__class__.__name__ != 'Account':
            raise TypeError('account must be an Account instance, '
                            'not a %s' % type(account))

        if album is not None and album.__class__.__name__ != 'Album':
            raise TypeError('album must be an Album instance, '
                            'not a %s' % type(album))

        if artist is not None and artist.__class__.__name__ != 'Artist':
            raise TypeError('artist must be an Artist instance, '
                            'not a %s' % type(account))

        if min_post_time is not None and type(min_post_time) is not datetime:
            raise TypeError('min_post_time must be a datetime.datetime, '
                            'not a %s' % type(min_post_time))

        if max_post_time is not None and type(max_post_time) is not datetime:
            raise TypeError('max_post_time must be a datetime.datetime, '
                            'not a %s' % type(max_post_time))

        filter_kwargs = {}

        if account:
            filter_kwargs['scrobbles__account'] = account

        if album:
            filter_kwargs['scrobbles__album'] = album

        if artist:
            filter_kwargs['scrobbles__artist'] = artist

        if track:
            filter_kwargs['scrobbles__track'] = track

        if min_post_time and max_post_time:
            filter_kwargs['scrobbles__post_time__gte'] = min_post_time
            filter_kwargs['scrobbles__post_time__lte'] = max_post_time

        elif min_post_time:
            filter_kwargs['scrobbles__post_time__gte'] = min_post_time

        elif max_post_time:
            filter_kwargs['scrobbles__post_time__lte'] = max_post_time

        qs = self.filter(**filter_kwargs)

        return qs.annotate(
                scrobble_count = models.Count('scrobbles', distinct=True)
            ).order_by('-scrobble_count')


class TracksManager(WithScrobbleCountsManager):
    """
    Adds a `scrobble_count` field to the Track objects.
    See WithScrobbleCountsManager for docs.
    """
    # We can't filter a list of Tracks by Tracks.
    is_filterable_by_track = False

    def with_scrobble_counts(self, **kwargs):
        "Pre-fetch all the Tracks' Artists."
        qs = super(TracksManager, self)\
                    .with_scrobble_counts(**kwargs)\
                    .prefetch_related('artist')
        return qs


class AlbumsManager(WithScrobbleCountsManager):
    """
    Adds a `scrobble_count` field to the Album objects.
    See WithScrobbleCountsManager for docs.
    """
    # We can't filter a list of Albums by Album.
    is_filterable_by_album = False

    def with_scrobble_counts(self, **kwargs):
        "Pre-fetch all the Albums' Artists."
        qs = super(AlbumsManager, self)\
                    .with_scrobble_counts(**kwargs)\
                    .prefetch_related('artist')
        return qs


class ArtistsManager(WithScrobbleCountsManager):
    """
    Adds a `scrobble_count` field to the Artist objects.
    See WithScrobbleCountsManager for docs.
    """
    # We can't filter a list of Artists by Artist.
    is_filterable_by_artist = False




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pinboard', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='bookmark',
            name='shared',
        ),
        migrations.AlterField(
            model_name='bookmark',
            name='raw',
            field=models.TextField(help_text=b'eg, the raw JSON from the API.', blank=True),
        ),
    ]



import calendar
from datetime import datetime, timedelta
import json
import pytz
import requests
import time
import urllib

from ditto import TITLE, VERSION
from .models import Account, Album, Artist, Scrobble, Track
from .utils import slugify_name
from ..core.utils import datetime_now


LASTFM_API_ENDPOINT = 'http://ws.audioscrobbler.com/2.0/'


class FetchError(Exception):
    pass


class ScrobblesFetcher(object):
    """
    Fetches scrobbles from the API for one Account.

    Use like one of:
        fetcher = ScrobblesFetcher(account)

    And then one of these ('recent' is the default):
        results = fetcher.fetch(fetch_type='recent')
        results = fetcher.fetch(fetch_type='all')
        results = fetcher.fetch(fetch_type='days', days=3)
    """
    # How many scrobbles do we fetch per page of results?
    items_per_page = 200

    def __init__(self, account):

        # Will be an Account object, passed into init()
        self.account = None

        # We'll set this to a datetime if we're fetching scrobbles since x.
        self.min_datetime = None

        self.page_number = 1

        self.total_pages = 1

        self.results_count = 0

        # What we'll return:
        self.return_value = {'fetched': 0}

        if isinstance(account, Account):
            self.return_value['account'] = str(account)
        else:
            raise ValueError("An Account object is required")

        if account.has_credentials():
            self.account = account
        else:
            self.return_value['success'] = False
            self.return_value['messages'] = ['Account has no API credentials']

    def fetch(self, fetch_type='recent', days=None):
        """
        Fetch and save scrobbles.

        Keyword arguments:
        fetch_type -- 'all', 'days' or 'recent'. The latter will fetch
                      scrobbles since the most recent Scrobble we already have.
        days -- if fetch_type is 'days', this should be an integer.

        Returns a dict like:
            {'success': True, 'account': 'gyford', 'fetched': 47,}
        Or:
            {'success': False, 'account': 'gyford', 'messages': ['Oops..',],}
        """

        if self.account and self.account.is_active == False:
            self.return_value['success'] = False
            self.return_value['messages'] = [
                    'The Account %s is currently marked as inactive.' %\
                                                        self.account.username]
            return self.return_value

        valid_fetch_types = ['all', 'days', 'recent']
        if fetch_type not in valid_fetch_types:
            raise ValueError('fetch_type should be one of %s' %\
                                                ', '.join(valid_fetch_types))

        if fetch_type == 'days':
            try:
                test = days + 1
            except TypeError:
                raise ValueError('days argument should be an integer')

            self.min_datetime = datetime_now() - timedelta(days=days)

        elif fetch_type == 'recent':
            try:
                scrobble = Scrobble.objects.latest('post_time')
                self.min_datetime = scrobble.post_time
            except Scrobble.DoesNotExist:
                pass

        self._fetch_pages()

        if self._not_failed():
            self.return_value['success'] = True
            self.return_value['fetched'] = self.results_count

        return self.return_value

    def _fetch_pages(self):
        while self.page_number <= self.total_pages and self._not_failed():
            self._fetch_page()
            self.page_number += 1
            time.sleep(0.5)

    def _fetch_page(self):
        """
        Fetch a single page of results.
        Uses the value of self.page_number.
        """
        fetch_time = datetime_now()

        try:
            results = self._send_request()
        except FetchError as e:
            self.return_value['success'] = False
            self.return_value['messages'] = [str(e)]
            return

        for scrobble in results:
            if 'date' in scrobble:
                # Don't save nowplaying scrobbles, that have no 'date'.
                self._save_scrobble(scrobble, fetch_time)
                self.results_count += 1

        return

    def _not_failed(self):
        """Has everything gone smoothly so far? ie, no failure registered?"""
        if 'success' not in self.return_value or self.return_value['success'] == True:
            return True
        else:
            return False

    def _api_method(self):
        "The name of the API method."
        return 'user.getrecenttracks'

    def _api_args(self):
        "Returns a dict of args for the API call."
        args = {
            'user':     self.account.username,
            'api_key':  self.account.api_key,
            'format':   'json',
            'method':   self._api_method(),
            'page':     self.page_number,
            'limit':    self.items_per_page,
        }

        if self.min_datetime:
            # Turn our datetime object into a unix timestamp:
            args['from'] = calendar.timegm(self.min_datetime.timetuple())

        return args

    def _send_request(self):
        """
        Send a request to the Last.fm API.

        Raises FetchError if something goes wrong.
        Returns a list of results if all goes well.
        """
        query_string = urllib.parse.urlencode(self._api_args())

        url = "{}?{}".format(LASTFM_API_ENDPOINT, query_string)

        try:
            response = requests.get(url,
                        headers={'User-Agent': 'Mozilla/5.0 (%s v%s)' % (
                                                            TITLE, VERSION)})
            response.raise_for_status() # Raises an exception on HTTP error.
        except requests.exceptions.RequestException as e:
            raise FetchError(
                    "Error when fetching Scrobbles (page %s): %s" % \
                                                    (self.page_number, str(e)))

        results = json.loads(response.text)

        if 'error' in results:
            raise FetchError(
                    "Error %s when fetching Scrobbles (page %s): %s" % \
                    (results['error'], self.page_number, results['message'])
                )

        # Set total number of pages first time round:
        attr = results['recenttracks']['@attr']
        if self.page_number == 1 and 'totalPages' in attr:
            self.total_pages = int(attr['totalPages'])

        return results['recenttracks']['track']

    def _save_scrobble(self, scrobble, fetch_time):
        """
        Saves/updates a scrobble.

        Arguments:
        scrobble -- A dict of data from the Last.fm API.
        fetch_time -- Datetime of when the data was fetched.
        """
        artist_slug, track_slug = self._get_slugs(scrobble['url'])

        artist, created = Artist.objects.update_or_create(
            slug=artist_slug.lower(),
            defaults={
                'name': scrobble['artist']['#text'],
                'original_slug': artist_slug,
                'mbid': scrobble['artist']['mbid'], # Might be "".
            }
        )

        track, created = Track.objects.update_or_create(
            slug=track_slug.lower(),
            artist=artist,
            defaults={
                'name': scrobble['name'],
                'original_slug': track_slug,
                'mbid': scrobble['mbid'], # Might be "".
            }
        )

        if scrobble['album']['#text'] == '':
            album = None
        else:
            # The API data doesn't provide a URL/slug for the album, so
            # we make our own:
            album_slug = slugify_name(scrobble['album']['#text'])

            album, created = Album.objects.update_or_create(
                slug=album_slug.lower(),
                artist=artist,
                defaults={
                    'name': scrobble['album']['#text'],
                    'original_slug': album_slug,
                    'mbid': scrobble['album']['mbid'], # Might be "".
                }
            )

        # Unixtime to datetime object:
        scrobble_time = datetime.utcfromtimestamp(
                            int(scrobble['date']['uts'])
                        ).replace(tzinfo=pytz.utc)

        scrobble_obj, created = Scrobble.objects.update_or_create(
            account=self.account,
            track=track,
            post_time=scrobble_time,
            defaults={
                'artist':       artist,
                'raw':          json.dumps(scrobble),
                'fetch_time':   fetch_time,
                'album':        album,
            }
        )

        return scrobble_obj

    def _get_slugs(self, scrobble_url):
        """
        Get the artist and track slugs from a scrobble's URL.
        The scrobble's URL is also the Track's URL.

        scrobble_url is like 'https://www.last.fm/music/Artist/_/Track'
        returns two strings, artist_slug and track_slug.
        """
        url = scrobble_url.rstrip('/')

        # Need to replace semicolons as urlparse() treats them (legitimately)
        # as alternatives to '&' as a query string separator, and so omits
        # anything after them.
        url = url.replace(';', '%3B')

        # www.last.fm/music/Artist/_/Track':
        url_path = urllib.parse.urlparse(url).path
        path_parts = url_path.split('/')

        artist_slug = path_parts[-3]  # 'Artist'
        track_slug = path_parts[-1]   # 'Track'

        # Put those naughty semicolons back in:
        artist_slug = artist_slug.replace('%3B', ';')
        track_slug = track_slug.replace('%3B', ';')

        return artist_slug, track_slug


class ScrobblesMultiAccountFetcher(object):
    """
    For fetching Scrobbles for ALL or ONE account(s).

    Usage example:
        results = ScrobblesMultiAccountFetcher().fetch(fetch_type='recent')

    Or:
        results = ScrobblesMultiAccountFetcher(username='bob').fetch(fetch_type='recent')

    results will be a list of dicts containing info about what was fetched (or
    went wrong) for each account.
    """

    # Will be a list of Account objects.
    accounts = []

    def __init__(self, username=None):
        """
        Gets all of the Accounts, or the single Account specified.

        username -- If username is set, we only use that Account, if active.
                    If it's not set, we use all active Accounts.
        """
        self.return_value = []

        if username is None:
            # Get all active Accounts.
            self.accounts = list(Account.objects.filter(is_active=True))
            if len(self.accounts) == 0:
                raise FetchError("No active Accounts were found to fetch.")
        else:
            # Find the Account associated with username.
            try:
                account = Account.objects.get(username=username)
            except Account.DoesNotExist:
                raise FetchError(
                    "There is no Account with the username '%s'" % username)
            if account.is_active == False:
                raise FetchError(
                    "The Account with the username '%s' is marked as inactive." % username)
            self.accounts = [account]

    def fetch(self, **kwargs):
        for account in self.accounts:
            self.return_value.append(
                ScrobblesFetcher(account).fetch(**kwargs)
            )

        return self.return_value




from datetime import datetime
from django import template
from django.urls import reverse
from django.http import QueryDict
from django.utils.html import format_html

from ..apps import ditto_apps


register = template.Library()


@register.simple_tag
def get_enabled_apps():
    """
    Returns a list of strings indicating which Ditto apps are enabled.
    e.g.
        ['flickr', 'twitter',]
    """
    return ditto_apps.enabled()


@register.simple_tag(takes_context=True)
def query_string(context, key, value):
    """
    For adding/replacing a key=value pair to the GET string for a URL.

    eg, if we're viewing ?p=3 and we do {% url_replace order 'taken' %}
    then this returns "p=3&order=taken"

    And, if we're viewing ?p=3&order=uploaded and we do the same thing, we get
    the same result (ie, the existing "order=uploaded" is replaced).

    Expects the request object in context to do the above; otherwise it will
    just return a query string with the supplied key=value pair.
    """
    try:
        request = context['request']
        args = request.GET.copy()
    except KeyError:
        args = QueryDict('').copy()
    args[key] = value
    return args.urlencode()


@register.simple_tag
def width_height(w, h, max_w, max_h):
    """Returns a string like:
        width="200" height="150"
    with the values limited to max_w and max_h, and scaled appropriately.

    If both w and h are smaller than their maximums, they're returned as is.
    """
    ratio = 1

    if w > max_w and h > max_h:
        ratio = min( (max_w / w), (max_h / h) )

    elif w > max_w:
        ratio = max_w / w

    elif h > max_h:
        ratio = max_h / h

    width = int(round(w * ratio))
    height = int(round(h * ratio))

    return format_html('width="%s" height="%s"' % (width, height))


@register.simple_tag
def display_time(dt, link_to_day=False, granularity=0, case=None):
    """Return the HTML to display the time a Photo, Tweet, etc.

    dt -- The datetime.
    view -- Nothing or 'detail' or 'day', probably.
    granularity -- A number indicating how detailed the datetime is, based on
                    https://www.flickr.com/services/api/misc.dates.html
    case -- How the visible text will be treated. By default will be lowercase
                except for Month names. None, 'lower', or 'capfirst'.

    For a 'day' view, just returns the date/time as text.
    For other views returns it including a link to the ditto:day_archive page
        for that date.
    Both wrapped in a <time> tag.

    See also http://www.brucelawson.co.uk/2012/best-of-time/ for <time> tag.
    """

    if granularity == 8:
        visible_time = 'circa %s' % dt.strftime('%Y')
        stamp = dt.strftime('%Y')

    elif granularity == 6:
        visible_time = 'sometime in %s' % dt.strftime('%Y')
        stamp = dt.strftime('%Y')

    elif granularity == 4:
        visible_time = 'sometime in %s' % dt.strftime('%b&nbsp;%Y')
        stamp = dt.strftime('%Y-%m')

    else:
        # Exact time and date.
        # We can link to the date, if link_to_day=True.

        stamp = dt.strftime('%Y-%m-%d %H:%M:%S')

        # The date and time formats for display:
        d_fmt = '%-d&nbsp;%b&nbsp;%Y'
        t_fmt = '%H:%M'

        if link_to_day:
            url = reverse('ditto:day_archive', kwargs={
                        'year':     dt.strftime('%Y'),
                        'month':    dt.strftime('%m'),
                        'day':      dt.strftime('%d'),
                    })

            visible_time = '%(time)s on <a href="%(url)s" title="All items from this day">%(date)s</a>' % {
                    'time': dt.strftime(t_fmt),
                    'url': url,
                    'date': dt.strftime(d_fmt),
                }
        else:
            visible_time = dt.strftime(t_fmt + ' on ' + d_fmt)

    if case == 'lower':
        visible_time = visible_time.lower()
    elif case == 'capfirst':
        visible_time = visible_time[0].upper() + visible_time[1:]

    return format_html('<time datetime="%(stamp)s">%(visible)s</time>' % {
                'stamp': stamp,
                'visible': visible_time
            })


@register.simple_tag(takes_context=True)
def current_url_name(context):
    """
    Returns the name of the current URL, namespaced, or False.

    Example usage:

        {% current_url_name as url_name %}

        <a href="#"{% if url_name == 'myapp:home' %} class="active"{% endif %}">Home</a>

    """
    url_name = False
    if context.request.resolver_match:
        url_name = "{}:{}".format(
                                context.request.resolver_match.namespace,
                                context.request.resolver_match.url_name
                            )
    return url_name



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pinboard', '0008_bookmark_tags'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bookmark',
            name='post_time',
            field=models.DateTimeField(help_text=b'The time this was created on Pinboard.', null=True, blank=True),
        ),
    ]



# coding: utf-8
from django.utils.html import strip_tags
from django.utils.text import Truncator

import datetime
import pytz

from django.db.models import Count


def truncate_string(text, strip_html=True, chars=255, truncate=u'…', at_word_boundary=False):
    """Truncate a string to a certain length, removing line breaks and mutliple
    spaces, optionally removing HTML, and appending a 'truncate' string.

    Keyword arguments:
    strip_html -- boolean.
    chars -- Number of characters to return.
    at_word_boundary -- Only truncate at a word boundary, which will probably
        result in a string shorter than chars.
    truncate -- String to add to the end.
    """
    if strip_html:
        text = strip_tags(text)
    text = text.replace('\n', ' ').replace('\r', '')
    text = ' '.join(text.split())
    if at_word_boundary:
        if len(text) > chars:
            text = text[:chars].rsplit(' ', 1)[0] + truncate
    else:
        text = Truncator(text).chars(chars, html=False, truncate=truncate)
    return text


def datetime_now():
    """Just returns a datetime object for now in UTC, with UTC timezone.
    Because I was doing this a lot in various places.
    """
    return datetime.datetime.utcnow().replace(tzinfo=pytz.utc)


def datetime_from_str(s):
    """A shortcut for making a UTC datetime from a string like
    '2015-08-11 12:00:00'.
    """
    return datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S').replace(
                                                            tzinfo=pytz.utc)

def get_annual_item_counts(qs, field_name='post_year'):
    """
    Takes a QuerySet, probably of a DittoItem child class like Photo or Tweet,
    and returns a list of dicts with 'year' and 'count' keys. eg:
        [
            {'year': 2015, 'count': 123},
            {'year': 2016, 'count': 456},
            {'year': 2017, 'count': 0},
            {'year': 2018, 'count': 789},
        ]

    Note, the first query we do here gets a QuerySet that only includes years
    in which there are any counts:
        [
            {'post_year': 2015, 'count': 123},
            {'post_year': 2016, 'count': 456},
            # Nothing included for 2017, which has count 0.
            {'post_year': 2018, 'count': 789},
        ]

    So the rest of this function is all about filling in those gaps.

    Used in all the annual_[thing]_counts() template tags.

    Arguments:
        qs -- The QuerySet.
        field_name -- Name of the key identifying the years.
    """

    qs = qs.values(field_name)\
                    .annotate(count=Count('id'))\
                    .values(field_name, 'count')\
                    .order_by(field_name)

    # Just in case. eg, trying to get counts for a private Twitter account:
    if len(qs) == 0:
        return []

    # Make a set of years like {2015, 2016, 2018}:
    years_with_counts = set(y[field_name] for y in qs)

    # Make a set of years with no gaps like {2015, 2016, 2017, 2018}:
    all_years = sorted(set(
                    range(min(years_with_counts), max(years_with_counts) + 1)
                ))

    # Translate original qs into {'2015': 123, '2016': 456, '2018': 789}:
    qs_dict = {}
    for row in qs:
        qs_dict[ str(row[field_name]) ] = row['count']

    # Make the final list of dicts.
    # An element for every year between min and max, even if it had no items,
    # in which case its count is 0.
    results = []
    for y in all_years:
        try:
            count = qs_dict[str(y)]
        except KeyError:
            count = 0
        results.append( {'year': y, 'count': count} )

    return results




import argparse

from django.core.management.base import CommandError

from ....core.management.commands import DittoBaseCommand
from ...fetch import ScrobblesMultiAccountFetcher


class Command(DittoBaseCommand):

    # What we're fetching:
    singular_noun = 'Scrobble'
    plural_noun = 'Scrobbles'

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--account',
            action='store',
            default=False,
            help='The username of the Last.fm user to fetch Scrobbles for. e.g. "rj".'
        )

        parser.add_argument(
            '--days',
            action='store',
            default=False,
            help='Number of days to fetch. Or "all" for ALL scrobbles. Or omit to only fetch since the most recently fetched scrobble.'
        )

    def handle(self, *args, **options):
        # We might be fetching for a specific account or all (None).
        username = options['account'] if options['account'] else None

        # Default:
        fetch_type = 'recent'

        if options['days']:
            # Will be either 'all' or a number; make the number an int.
            if options['days'].isdigit():
                options['days'] = int(options['days'])
                fetch_type = 'days'

            elif options['days'] == 'all':
                fetch_type = 'all'

            else:
                raise CommandError("--days should be an integer or 'all'.")

        fetcher = ScrobblesMultiAccountFetcher(username=username)

        if fetch_type == 'days':
            results = fetcher.fetch(fetch_type=fetch_type, days=options['days'])
        else:
            results = fetcher.fetch(fetch_type=fetch_type)

        self.output_results(results, options.get('verbosity', 1))

